'''Module with classes and methods to analyse and process exported geomodel grids
Created on 21/03/2014
@author: Florian Wellmann (some parts originally developed by Erik Schaeffer)
'''
import numpy as np
#import pynoddy
import subprocess
import os.path
import platform

try:
    import matplotlib.pyplot as plt
except ImportError:
    print("\n\n\tMatplotlib not installed - plotting functions will not work!\n\n\n")

# import mpl_toolkits
# from matplotlib.ticker import MultipleLocator, FormatStrFormatter

# to convert python variable types to cpp types
import ctypes
# to create array
from numpy.ctypeslib import ndpointer
# to create folder
import os

import pyvista as pv
# read out and change xml file (here only used to read out model boundary information)
import geomodeller_xml_obj as GO


class GeoGrid():
    """Object definition for exported geomodel grids"""

    def __init__(self, **kwds):
        """GeoGrid contains methods to load, analyse, and process exported geomodel grids
        **Optional Keywords**:
            - *grid_filename* = string : filename of exported grid
            - *delxyz_filename* = string : file with model discretisation
            - *dimensions_filename* = string : file with model dimension (coordinates)
        """

        if 'grid_filename' in kwds:
            self.grid_filename = kwds['grid_filename']
        if 'delxyz_filename' in kwds:
            self.delxyz_filename = kwds['delxyz_filename']
        if 'dimensions_filename' in kwds:
            self.dimensions_filename = kwds['dimensions_filename']

    def __add__(self, G_other):
        """Combine grid with another GeoGrid if regions are overlapping"""
        # check overlap
        print (self.ymin, self.ymax)
        print (G_other.ymin, G_other.ymax)
        if (G_other.ymin < self.ymax and G_other.ymin > self.ymin):
            print("Grids overlapping in y-direction between %.0f and %.0f" %
                  (G_other.ymin, self.ymax))

    def load_grid(self):
        """Load exported grid, discretisation and dimensions from file"""
        if not hasattr(self, 'grid_filename'):
            raise AttributeError("Grid filename is not defined!")
        self.grid = np.loadtxt(self.grid_filename,
                               delimiter = ',',
                               dtype='int',
                               unpack=False)
        if hasattr(self, 'delxyz_filename'):
            self.load_delxyz(self.delxyz_filename)
            self.adjust_gridshape()
        if hasattr(self, 'dimensions_filename'):
            self.load_dimensions(self.dimensions_filename)

    def load_delxyz(self, delxyz_filename):
        """Load grid discretisation from file"""
        del_lines = open(delxyz_filename, 'r').readlines()
        d0 = del_lines[0].split("*")
        self.delx = np.array([float(d0[1]) for _ in range(int(d0[0]))])
        d1 = del_lines[1].split("*")
        self.dely = np.array([float(d1[1]) for _ in range(int(d1[0]))])
        d2 = del_lines[2].split(",")[:-1]
        self.delz = np.array([float(d) for d in d2])
        (self.nx, self.ny, self.nz) = (len(self.delx), len(self.dely), len(self.delz))
        (self.extent_x, self.extent_y, self.extent_z) = (sum(self.delx), sum(self.dely), sum(self.delz))

    def set_delxyz(self, delxyz):
        """Set delx, dely, delz arrays explicitly and update additional attributes
        **Arguments**:
            - *delxyz* = (delx-array, dely-array, delz-array): arrays with cell dimensions
        """
        self.delx, self.dely, self.delz = delxyz
        (self.nx, self.ny, self.nz) = (len(self.delx), len(self.dely), len(self.delz))
        (self.extent_x, self.extent_y, self.extent_z) = (sum(self.delx), sum(self.dely), sum(self.delz))

    def set_basename(self, name):
        """Set basename for grid exports, etc.
        **Arguments**:
            - *name* = string: basename
        """
        self.basename = name

    def load_dimensions(self, dimensions_filename):
        """Load project dimensions from file"""
        dim = [float(d) for d in open(dimensions_filename, 'r').readlines()[1].split(",")]
        (self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax) = dim
        # calculate cell centre positions in real world coordinates

    def define_regular_grid(self, nx, ny, nz):
        """Define a regular grid from defined project boundaries and given discretisations"""
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.delx = np.ones(nx) * (self.xmax - self.xmin) / nx
        self.dely = np.ones(ny) * (self.ymax - self.ymin) / ny
        self.delz = np.ones(nz) * (self.zmax - self.zmin) / nz
        # create (empty) grid object
        self.grid = np.ndarray((nx, ny, nz))
        # update model extent
        (self.extent_x, self.extent_y, self.extent_z) = (sum(self.delx), sum(self.dely), sum(self.delz))


    def define_irregular_grid(self, delx, dely, delz):
        """Set irregular grid according to delimter arrays in each direction"""
        self.delx = delx
        self.dely = dely
        self.delz = delz
        self.nx = len(delx)
        self.ny = len(dely)
        self.nz = len(delz)
        # create (empty) grid object
        self.grid = np.ndarray((self.nx, self.ny, self.nz))
        # update model extent
        (self.extent_x, self.extent_y, self.extent_z) = (sum(self.delx), sum(self.dely), sum(self.delz))

    def get_dimensions_from_geomodeller_xml_project(self, xml_filename):
        """Get grid dimensions from Geomodeller project
        **Arguments**:
            - *xml_filename* = string: filename of Geomodeller XML file
        """
        # Note: this implementation is based on the Geomodeller API
        # The boundaries could theoretically also be extracted from the XML file
        # directly, e.g. using the geomodeller_xml_obj module - but this would
        # require an additional module being loaded, so avoid here!
        filename_ctypes = ctypes.c_char_p(xml_filename)
        # get model boundaries
            #Detection of operative system:
        if platform.system() == "Linux":
            lib = ctypes.CDLL('./libgeomod.so') #linux
        elif platform.system() == "Windows":
            lib = ctypes.windll.LoadLibrary(os.path.dirname(os.path.abspath(__file__)) + os.path.sep + "libgeomodwin_leak6.dll")     #windows
        else:
            print("Your operative system is not supported")
        lib.get_model_bounds.restype = ndpointer(dtype=ctypes.c_int, shape=(6,))
        boundaries = lib.get_model_bounds(filename_ctypes)
        (self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax) = boundaries
        self.extent_x = self.xmax - self.xmin
        self.extent_y = self.ymax - self.ymin
        self.extent_z = self.zmax - self.zmin


    def set_densities(self, densities):
        """Set layer densities
        **Arguments**:
            - *densities* = dictionary of floats: densities for geology ids
        """
        self.densities = densities

    def set_sus(self, sus):
        """Set layer susceptibilities
        **Arguments**:
            - *us* = dictionary of floats: magnetic susceptibilities for geology ids
        """
        self.sus = sus

    def set_dimensions(self, **kwds):
        """Set model dimensions, if no argument provided: xmin = 0, max = sum(delx) and accordingly for y,z
        **Optional keywords**:
            - *dim* = (xmin, xmax, ymin, ymax, zmin, zmax) : set dimensions explicitly
        """
        if "dim" in kwds:
            (self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax) = kwds['dim']
        else:
            self.xmin, self.ymin, self.zmin = (0., 0., 0.)
            self.xmax, self.ymax, self.zmax = (sum(self.delx), sum(self.dely), sum(self.delz))

    def determine_cell_centers(self):
        """Determine cell centers for all coordinate directions in "real-world" coordinates"""
        if not hasattr(self, 'xmin'):
            raise AttributeError("Please define grid dimensions first")
        sum_delx = np.cumsum(self.delx)
        sum_dely = np.cumsum(self.dely)
        sum_delz = np.cumsum(self.delz)
        self.cell_centers_x = np.array([sum_delx[i] - self.delx[i] / 2. for i in range(self.nx)]) + self.xmin
        self.cell_centers_y = np.array([sum_dely[i] - self.dely[i] / 2. for i in range(self.ny)]) + self.ymin
        self.cell_centers_z = np.array([sum_delz[i] - self.delz[i] / 2. for i in range(self.nz)]) + self.zmin

    def determine_cell_boundaries(self):
        """Determine cell boundaries for all coordinates in "real-world" coordinates"""
        if not hasattr(self, 'xmin'):
            raise AttributeError("Please define grid dimensions first")
        sum_delx = np.cumsum(self.delx)
        sum_dely = np.cumsum(self.dely)
        sum_delz = np.cumsum(self.delz)
        self.boundaries_x = np.ndarray((self.nx+1))
        self.boundaries_x[0] = 0
        self.boundaries_x[1:] = sum_delx
        self.boundaries_y = np.ndarray((self.ny+1))
        self.boundaries_y[0] = 0
        self.boundaries_y[1:] = sum_dely
        self.boundaries_z = np.ndarray((self.nz+1))
        self.boundaries_z[0] = 0
        self.boundaries_z[1:] = sum_delz

        # create a list with all bounds
        self.bounds = [self.boundaries_y[0], self.boundaries_y[-1],
                       self.boundaries_x[0], self.boundaries_x[-1],
                       self.boundaries_z[0], self.boundaries_z[-1]]


    def adjust_gridshape(self):
        """Reshape numpy array to reflect model dimensions"""
        self.grid = np.reshape(self.grid, (self.nz, self.ny, self.nx))
        self.grid = np.swapaxes(self.grid, 0, 2)
        # self.grid = np.swapaxes(self.grid, 0, 1)


    def export_to_vtk(self, vtk_filename="geo_grid.vtr", real_coords = True, **kwds):
        """Export grid to VTK for visualisation
        **Arguments**:
            - *vtk_filename* = string : vtk filename (obviously...)
            - *real_coords* = bool : model extent in "real world" coordinates
        **Optional Keywords**:
            - *grid* = numpy grid : grid to save to vtk (default: self.grid)
            - *var_name* = string : name of variable to plot (default: Geology)
        """
        grid = kwds.get("grid", self.grid)
        var_name = kwds.get("var_name", "Geology")
        # define coordinates
        x = np.zeros(self.nx + 1)
        y = np.zeros(self.ny + 1)
        z = np.zeros(self.nz + 1)
        x[1:] = np.cumsum(self.delx)
        y[1:] = np.cumsum(self.dely)
        z[1:] = np.cumsum(self.delz)

        # plot in coordinates
        if real_coords:
            x += self.xmin
            y += self.ymin
            z += self.zmin
            
        out = pv.RectilinearGrid(x,y,z)
        out['var_name'] = grid
        out.save(vtk_filename)

    def export_to_csv(self, filename = "geo_grid.csv"):
        """Export grid to x,y,z,value pairs in a csv file
        Ordering is x-dominant (first increase in x, then y, then z)
        **Arguments**:
            - *filename* = string : filename of csv file (default: geo_grid.csv)
        """
        f = open(filename, 'w')
        for zz in self.delz:
            for yy in self.dely:
                for xx in self.delx:
                    f.write("%.1f,%.1f,%.1f,%.d" % (xx,yy,zz,self.grid[xx,yy,zz]))
        f.close()


    def determine_geology_ids(self):
        """Determine all ids assigned to cells in the grid"""
        self.unit_ids = np.unique(self.grid)

    def get_name_mapping_from_file(self, filename):
        """Get the mapping between unit_ids in the model and real geological names
        from a csv file (e.g. the SHEMAT property file)
        **Arguments**:
            - *filename* = string : filename of csv file with id, name entries
        """
        self.unit_name = {}
        filelines = open(filename, 'r').readlines()[1:]
        for line in filelines:
            l = line.split(",")
            self.unit_name[int(l[1])] = l[0]

    def get_name_mapping_from_dict(self, unit_name_dict):
        """Get the name mapping directly from a dictionary
        **Arguments**:
            - *unit_name_dict* = dict with "name" : unit_id (int) pairs
        """
        self.unit_name = unit_name_dict


    def remap_ids(self, mapping_dictionary):
        """Remap geological unit ids to new ids as defined in mapping dictionary
        **Arguments**:
            - *mapping_dictionary* = dict : {1 : 1, 2 : 3, ...} : e.g.: retain
            id 1, but map id 2 to 3 (note: if id not specified, it will be retained)
        """
        # first step: create a single mesh for each id to avoid accidential
        # overwriting below (there might be a better solution...)
        if not hasattr(self, 'unit_ids'):
            self.determine_geology_ids()
        geol_grid_ind = {}
        for k,v in mapping_dictionary.items():
            geol_grid_ind[k] = self.grid == k
            print("Remap id %d -> %d" % (k,v))
        # now reassign values in actual grid
        for k,v in mapping_dictionary.items():
            print("Reassign id %d to grid" % v)
            self.grid[geol_grid_ind[k]] = v
        # update global geology ids
        self.determine_geology_ids()

    def determine_cell_volumes(self):
        """Determine cell volumes for each cell (e.g. for total formation volume calculation)"""
        self.cell_volume = np.ndarray(np.shape(self.grid))
        for k,dz in enumerate(self.delz):
            for j,dy in enumerate(self.dely):
                for i,dx in enumerate(self.delx):
                    self.cell_volume[i,j,k] = dx * dy * dz


    def determine_indicator_grids(self):
        """Determine indicator grids for all geological units"""
        self.indicator_grids = {}
        if not hasattr(self, 'unit_ids'):
            self.determine_geology_ids()
        grid_ones = np.ones(np.shape(self.grid))
        for unit_id in self.unit_ids:
            self.indicator_grids[unit_id] = grid_ones * (self.grid == unit_id)

    def determine_id_volumes(self):
        """Determine the total volume of each unit id in the grid
        (for example for cell discretisation studies, etc."""
        if not hasattr(self, 'cell_volume'):
            self.determine_cell_volumes()
        if not hasattr(self, 'indicator_grids'):
            self.determine_indicator_grids()
        self.id_volumes = {}
        for unit_id in self.unit_ids:
            self.id_volumes[unit_id] = np.sum(self.indicator_grids[unit_id] * self.cell_volume)

    def print_unit_names_volumes(self):
        """Formatted output to STDOUT of unit names (or ids, if names are note
        defined) and calculated volumes
        """
        if not hasattr(self, 'id_vikumes'):
            self.determine_id_volumes()

        if hasattr(self, "unit_name"):
            # print with real geological names
            print("Total volumes of modelled geological units:\n")
            for unit_id in self.unit_ids:
                print("%26s : %.2f km^3" % (self.unit_name[unit_id],
                                            self.id_volumes[unit_id]/1E9))
        else:
            # print with unit ids only
            print("Total volumes of modelled geological units:\n")
            for unit_id in self.unit_ids:
                print("%3d : %.2f km^3" % (unit_id,
                                            self.id_volumes[unit_id]/1E9))


    def extract_subgrid(self, subrange, **kwds):
        """Extract a subgrid model from existing grid
        **Arguments**:
            - *subrange* = (x_from, x_to, y_from, y_to, z_from, z_to) : range for submodel in either cell or world coords
        **Optional keywords**:
            - *range_type* = 'cell', 'world' : define if subrange in cell ids (default) or real-world coordinates
        """
        range_type = kwds.get('range_type', 'cell')

        if not hasattr(self, 'boundaries_x'):
            self.determine_cell_boundaries()

        if range_type == 'world':
            # determine cells
            subrange[0] = np.argwhere(self.boundaries_x > subrange[0])[0][0]
            subrange[1] = np.argwhere(self.boundaries_x < subrange[1])[-1][0]
            subrange[2] = np.argwhere(self.boundaries_y > subrange[2])[0][0]
            subrange[3] = np.argwhere(self.boundaries_y < subrange[3])[-1][0]
            subrange[4] = np.argwhere(self.boundaries_z > subrange[4])[0][0]
            subrange[5] = np.argwhere(self.boundaries_z < subrange[5])[-1][0]

        # create a copy of the original grid
        import copy
        subgrid = copy.deepcopy(self)

        # extract grid
        subgrid.grid = self.grid[subrange[0]:subrange[1],
                                 subrange[2]:subrange[3],
                                 subrange[4]:subrange[5]]

        subgrid.nx = subrange[1] - subrange[0]
        subgrid.ny = subrange[3] - subrange[2]
        subgrid.nz = subrange[5] - subrange[4]

        # update extent
        subgrid.xmin = self.boundaries_x[subrange[0]]
        subgrid.xmax = self.boundaries_x[subrange[1]]
        subgrid.ymin = self.boundaries_y[subrange[2]]
        subgrid.ymax = self.boundaries_y[subrange[3]]
        subgrid.zmin = self.boundaries_z[subrange[4]]
        subgrid.zmax = self.boundaries_z[subrange[5]]

        subgrid.extent_x = subgrid.xmax - subgrid.xmin
        subgrid.extent_y = subgrid.ymax - subgrid.ymin
        subgrid.extent_z = subgrid.zmax - subgrid.zmin

        # update cell spacings
        subgrid.delx = self.delx[subrange[0]:subrange[1]]
        subgrid.dely = self.dely[subrange[2]:subrange[3]]
        subgrid.delz = self.delz[subrange[4]:subrange[5]]

        # now: update other attributes:
        subgrid.determine_cell_centers()
        subgrid.determine_cell_boundaries()
        subgrid.determine_cell_volumes()
        subgrid.determine_geology_ids()

        # finally: return subgrid
        return subgrid



# ******************************************************************************
#  Some additional helper functions
# ******************************************************************************

def combine_grids(G1, G2, direction, merge_type = 'keep_first', **kwds):
    """Combine two grids along one axis
    ..Note: this implementation assumes (for now) that the overlap is perfectly matching,
    i.e. grid cell sizes identical and at equal positions, or that they are perfectly adjacent!
    **Arguments**:
        - G1, G2 = GeoGrid : grids to be combined
        - direction = 'x', 'y', 'z': direction in which grids are combined
        - merge_type = method to combine grid:
            'keep_first' : keep elements of first grid (default)
            'keep_second' : keep elements of second grid
            'random' : randomly choose an element to retain
        ..Note: all other dimensions must be matching perfectly!!
    **Optional keywords**:
        - *overlap_analysis* = bool : perform a detailed analysis of the overlapping area, including
        mismatch. Also returns a second item, a GeoGrid with information on mismatch!
    **Returns**:
        - *G_comb* = GeoGrid with combined grid
        - *G_overlap* = Geogrid with analysis of overlap (of overlap_analysis=True)
    """
    overlap_analysis = kwds.get("overlap_analysis", False)
    # first step: determine overlap
    if direction == 'x':
        if G2.xmax > G1.xmax:
            overlap_min = G2.xmin
            overlap_max = G1.xmax
            # identifier alias for grids with higher/ lower values
            G_high = G2
            G_low = G1
        else:
            overlap_min = G1.xmin
            overlap_max = G2.xmax
            # identifier alias for grids with higher/ lower values
            G_high = G1
            G_low = G2

        # check if all other dimensions are perfectly matching
        if (G1.ymin != G2.ymin) or (G1.zmin != G2.zmin) or \
            (G1.ymax != G2.ymax) or (G1.zmax != G2.zmax):
            raise ValueError("Other dimensions (apart from %s) not perfectly matching! Check and try again!" % direction)

    elif direction == 'y':
        if G2.ymax > G1.ymax:
            overlap_min = G2.ymin
            overlap_max = G1.ymax
            # identifier alias for grids with higher/ lower values
            G_high = G2
            G_low = G1
        else:
            overlap_min = G1.ymin
            overlap_max = G2.ymax
            # identifier alias for grids with higher/ lower values
            G_high = G1
            G_low = G2

        # check if all other dimensions are perfectly matching
        if (G1.xmin != G2.xmin) or (G1.zmin != G2.zmin) or \
            (G1.xmax != G2.xmax) or (G1.zmax != G2.zmax):
            raise ValueError("Other dimensions (apart from %s) not perfectly matching! Check and try again!" % direction)

    elif direction == 'z':
        if G2.zmax > G1.zmax:
            overlap_min = G2.zmin
            overlap_max = G1.zmax
            # identifier alias for grids with higher/ lower values
            G_high = G2
            G_low = G1
        else:
            overlap_min = G1.zmin
            overlap_max = G2.zmax
            # identifier alias for grids with higher/ lower values
            G_high = G1
            G_low = G2

        # check if all other dimensions are perfectly matching
        if (G1.ymin != G2.ymin) or (G1.xmin != G2.xmin) or \
            (G1.ymax != G2.ymax) or (G1.xmax != G2.xmax):
            raise ValueError("Other dimensions (apart from %s) not perfectly matching! Check and try again!" % direction)

    overlap = overlap_max - overlap_min

    if overlap == 0:
        print("Grids perfectly adjacent")
    elif overlap < 0:
        raise ValueError("No overlap between grids! Check and try again!")
    else:
        print("Positive overlap in %s direction of %f meters" % (direction, overlap))

    # determine cell centers
    G1.determine_cell_centers()
    G2.determine_cell_centers()

    # intialise new grid
    G_comb = GeoGrid()
    # initialise overlap grid, if analyis performed
    if overlap_analysis:
        G_overlap = GeoGrid()

    if direction == 'x':
        pass
    elif direction == 'y':
        #=======================================================================
        # Perform overlap analysis
        #=======================================================================

        # initialise overlap grid with dimensions of overlap
        G_overlap.set_dimensions(dim = (G1.xmin, G1.xmax, overlap_min, overlap_max, G1.zmin, G1.zmax))
        G_low_ids = np.where(G_low.cell_centers_y > overlap_min)[0]
        G_high_ids = np.where(G_high.cell_centers_y < overlap_max)[0]
        delx = G1.delx
        dely = G_low.dely[G_low_ids]
        delz = G1.delz
        G_overlap.set_delxyz((delx, dely, delz))
        # check if overlap region is identical
        if not (len(G_low_ids) == len(G_high_ids)):
            raise ValueError("Overlap length not identical, please check and try again!")
        # now: determine overlap mismatch
        G_overlap.grid = G_low.grid[:,G_low_ids,:] - G_high.grid[:,G_high_ids,:]
        # for some very strange reason, this next step is necessary to enable the VTK
        G_overlap.grid = G_overlap.grid + np.zeros(G_overlap.grid.shape)
        #

        #=======================================================================
        # Set up combined grid
        #=======================================================================

        G_comb.set_dimensions(dim = (G1.xmin, G1.xmax, G_low.ymin, G_high.ymax, G1.zmin, G1.zmax))
        # combine dely arrays
        dely = np.hstack((G_low.dely[:G_low_ids[0]], G_high.dely))
        G_comb.set_delxyz((delx, dely, delz))

        #=======================================================================
        # Now merge grids
        #=======================================================================
        if merge_type == 'keep_first':
            if G1.ymax > G2.ymax:
                G_comb.grid = np.concatenate((G2._grid[:, :G_low_ids[0], :], G1._grid), axis=1)
            else:
                G_comb.grid = np.concatenate((G1._grid, G2._grid[:, :G_low_ids[0], :]), axis=1)

        elif merge_type == 'keep_second':
            pass
        elif merge_type == 'random':
            pass
        else:
            raise ValueError("Merge type %s not recognised! Please check and try again!" % merge_type)

    elif direction == 'z':
        pass




    # Return combined grid and results of overlap analysis, if determined
    if overlap_analysis:
        return G_comb, G_overlap
    else:
        return G_comb


def optimial_cell_increase(starting_cell_width, n_cells, width):
    """Determine an array with optimal cell width for a defined starting cell width,
    total number of cells, and total width
    Basically, this function optimised a factor between two cells to obtain a total
    width
    **Arguments**:
    	- *starting_cell_width* = float : width of starting/ inner cell
	- *n_cells* = int : total number of cells
	- *total_width* = float : total width (sum over all elements in array)
    **Returns**:
        del_array : numpy.ndarray with cell discretisations
    Note: optmisation with scipy.optimize - better (analytical?) methods might exist but
    I can't think of them at the moment
    """
    import scipy.optimize
    # define some helper functions

    def width_sum(inc_factor, inner_cell, n_cells, total_width):
        return sum(del_array(inc_factor, inner_cell, n_cells)) - total_width

    def del_array(inc_factor, inner_cell, n_cells):
        return np.array([inner_cell * inc_factor**i for i in range(n_cells)])

    # now the actual optimisation step:
    opti_factor = scipy.optimize.fsolve(width_sum, 1.1, (starting_cell_width, n_cells, width))

    # return the discretisation array
    return del_array(opti_factor, starting_cell_width, n_cells).flatten()



"""
This file is part of gempy.

Created on 07.08.2019

@author: Jan von Harten
"""

import warnings
try:
    from scipy.spatial.distance import cdist
except ImportError:
    warnings.warn('scipy.spatial package is not installed.')

import numpy as np
import pandas as pd
from gempy_legacy.plot import _visualization_2d, _plot, helpers
import matplotlib.cm as cm
import matplotlib.pyplot as plt
from copy import deepcopy

class domain(object):

    def __init__(self, model, domain=None, data=None, set_mean=None):

        # set model from a gempy solution
        # TODO: Check if I actually need all this or if its easier to just get grid and lith of the solution
        self.sol = model

        # set kriging surfaces, basically in which lithologies to do all this, default is everything
        # TODO: Maybe also allow to pass a gempy regular grid object
        if domain is None:
            domain = np.unique(self.sol.lith_block)
        self.set_domain(domain)

        # set data, default is None
        # TODO: need to figure out a way to then set mean and variance for the SGS and SK
        if data is None:
            data = None  # why do you do this, data is none already if it is none?
        self.set_data(data)

        # basic statistics of data
        # TODO: allow to set this  for SK ???
        if set_mean is None:
            set_mean = np.mean(data[:, 3])
        self.inp_mean = set_mean

        self.inp_var = np.var(data[:, 3])
        self.inp_std = np.sqrt(self.inp_var)

    def set_domain(self, domain):
        """
        Method to cut domain by array of surfaces. Simply masking the lith_block with array of input lithologies
        applying mask to grid.
        Args:
            domain (np.array)(x,) = array containing all surfaces of interest from the gempy model that
                                            the operation should be performed in
        Returns:
            ? Nothing cause of self - is this good practice?
            """
        # set domain to variable of class
        self.domain = domain

        # mask by array of input surfaces (by id, can be from different series)
        self.mask = np.isin(self.sol.lith_block, self.domain)

        # Apply mask to lith_block and grid
        self.krig_lith = self.sol.lith_block[self.mask]
        self.krig_grid = self.sol.grid.values[self.mask]

    def set_data(self, data):
        """
        Method to set input data from csv or numpy array.
        Args:
            data (np.array)(x,4)  = array of input data (conditioning) with [:,0]=x coordinate, [:,1]=y coordinate
                                            [:,2]=z coordinate and [:,3]=value of measured property
        Returns:
            ? Nothing cause of self - is this good practice?
            """
        # set domain to variable of class
        self.data = data

        # create dataframe of input data for calling
        d = {'X': data[:, 0], 'Y': data[:, 1], 'Z': data[:, 2], 'property': data[:, 3]}
        self.data_df = pd.DataFrame(data=d)


class variogram_model(object):

    # class containing all the variogram functionality

    def __init__(self, theoretical_model=None, range_=1, sill=1, nugget=0):

        if theoretical_model is None:
            theoretical_model = 'exponential'
        self.theoretical_model = theoretical_model

        # default
        self.range_ = range_
        self.sill = sill
        self.nugget = nugget

    def calculate_semivariance(self, d):

        if self.theoretical_model == 'exponential':
            gamma = self.exponential_variogram_model(d)
        elif self.theoretical_model == 'gaussian':
            gamma = self.gaussian_variogram_model(d)
        elif self.theoretical_model == 'spherical':
            gamma = self.spherical_variogram_model(d)
        else:
            print('theoretical varigoram model not understood')
        return gamma

    def calculate_covariance(self, d):

        if self.theoretical_model == 'exponential':
            gamma = self.exponential_covariance_model(d)
        elif self.theoretical_model == 'gaussian':
            gamma = self.gaussian_covariance_model(d)
        elif self.theoretical_model == 'spherical':
            gamma = self.spherical_covariance_model(d)
        else:
            print('theoretical varigoram model not understood')
        return gamma

    # TODO: Add more options
    # seems better now by changing psill in covariance model
    def exponential_variogram_model(self, d):
        '''Exponential variogram model, effective range approximately 3r, valid in R3'''
        psill = self.sill - self.nugget
        gamma = psill * (1. - np.exp(-(np.absolute(d) / (self.range_)))) + self.nugget
        return gamma

    def exponential_covariance_model(self, d):
        '''Exponential covariance model, effective range approximately 3r, valid in R3'''
        psill = self.sill - self.nugget
        cov = psill * (np.exp(-(np.absolute(d) / (self.range_))))
        return cov

    def gaussian_variogram_model(self, d):
        '''Gaussian variogram model, effective range approximately sqrt(3r),
        deprecated due to reverse curvature near orgin, valid in R3'''
        psill = self.sill - self.nugget
        gamma = psill * (1. - np.exp(-d ** 2. / (self.range_) ** 2.)) + self.nugget
        return gamma

    def gaussian_covariance_model(self, d):
        '''Gaussian covariance model, effective range approximately sqrt(3r),
        deprecated due to reverse curvature near orgin, valid in R3'''
        psill = self.sill - self.nugget
        gamma = psill * (np.exp(-d ** 2. / (self.range_) ** 2.))
        return gamma

    def spherical_variogram_model(self, d):
        '''Spherical variogram model, effective range equals range parameter, valid in R3'''
        psill = self.sill - self.nugget
        d = d.astype(float)
        gamma = np.piecewise(d, [d <= self.range_, d > self.range_],
                             [lambda d:
                              psill * ((3. * d) / (2. * self.range_)
                                       - (d ** 3.) / (2. * self.range_ ** 3.)) + self.nugget,
                              lambda d: self.sill])
        return gamma

    def spherical_covariance_model(self, d):
        '''Spherical covariance model, effective range equals range parameter, valid in R3'''
        psill = self.sill - self.nugget
        d = d.astype(float)
        gamma = np.piecewise(d, [d <= self.range_, d > self.range_],
                             [lambda d:
                              psill * (1 - ((3. * d) / (2. * self.range_)
                                            - (d ** 3.) / (2. * self.range_ ** 3.))),
                              lambda d: 0])
        return gamma

    # TODO: Make this better and nicer and everything
    # option for covariance
    # display range, sill, nugget, practical range etc.
    def plot(self, type_='variogram', show_parameters=True):

        if show_parameters == True:
            plt.axhline(self.sill, color='black', lw=1)
            plt.text(self.range_*2, self.sill, 'sill', fontsize=12, va='center', ha='center', backgroundcolor='w')
            plt.axvline(self.range_, color='black', lw=1)
            plt.text(self.range_, self.sill/2, 'range', fontsize=12, va='center', ha='center', backgroundcolor='w')

        if type_ == 'variogram':
            d = np.arange(0, self.range_*4, self.range_/1000)
            plt.plot(d, self.calculate_semivariance(d), label=self.theoretical_model + " variogram model")
            plt.ylabel('semivariance')
            plt.title('Variogram model')
            plt.legend()

        if type_ == 'covariance':
            d = np.arange(0, self.range_*4, self.range_/1000)
            plt.plot(d, self.calculate_covariance(d), label=self.theoretical_model + " covariance model")
            plt.ylabel('covariance')
            plt.title('Covariance model')
            plt.legend()

        if type_ == 'both':
            d = np.arange(0, self.range_*4, self.range_/1000)
            plt.plot(d, self.calculate_semivariance(d), label=self.theoretical_model + " variogram model")
            plt.plot(d, self.calculate_covariance(d), label=self.theoretical_model + " covariance model")
            plt.ylabel('semivariance/covariance')
            plt.title('Models of spatial correlation')
            plt.legend()

        plt.xlabel('lag distance')
        plt.ylim(0-self.sill/20, self.sill+self.sill/20)
        plt.xlim(0, self.range_*4)



class field_solution(object):

    def __init__(self, domain, variogram_model, results, field_type):

        self.results_df = results
        self.variogram_model = deepcopy(variogram_model)
        self.domain = deepcopy(domain)
        self.field_type = field_type

    def plot_results(self, geo_data, prop='val', direction='y', result='interpolation', cell_number=0, contour=False,
                     cmap='viridis', alpha=0, legend=False, interpolation='nearest', show_data=True):
        """
        TODO WRITE DOCSTRING
        Args:
            geo_data:
            prop: property that should be plotted - "val", "var" or "both"
            direction: x, y or z
            cell_number:
            contour:
            cmap:
            alpha:
            legend:

        Returns:

        """
        a = np.full_like(self.domain.mask, np.nan, dtype=np.double) #array like lith_block but with nan if outside domain

        est_vals = self.results_df['estimated value'].values
        est_var = self.results_df['estimation variance'].values

        # set values
        if prop == 'val':
            a[np.where(self.domain.mask == True)] = est_vals
        elif prop == 'var':
            a[np.where(self.domain.mask == True)] = est_var
        elif prop == 'both':
            a[np.where(self.domain.mask == True)] = est_vals
            b = np.full_like(self.domain.mask, np.nan, dtype=np.double)
            b[np.where(self.domain.mask == True)] = est_var
        else:
            print('prop must be val var or both')

        #create plot object
        p = _visualization_2d.PlotSolution(geo_data)
        _a, _b, _c, extent_val, x, y = p._slice(direction, cell_number)[:-2]

        #colors
        cmap = cm.get_cmap(cmap)
        cmap.set_bad(color='w', alpha=alpha) #define color and alpha for nan values

        # plot
        if prop is not 'both':
            if show_data:
                plt.scatter(self.domain.data_df[x].values, self.domain.data_df[y].values, marker='*', s=9, c='k')

            _plot.plot_section(geo_data, direction=direction, cell_number=cell_number)
            if contour == True:
                im = plt.contourf(a.reshape(self.domain.sol.grid.regular_grid.resolution)[_a, _b, _c].T, cmap=cmap,
                                  origin='lower', levels=25,
                                  extent=extent_val, interpolation=interpolation)
                if legend:
                    ax = plt.gca()
                    helpers.add_colorbar(axes=ax, label='prop', cs=im)
            else:
                im = plt.imshow(a.reshape(self.domain.sol.grid.regular_grid.resolution)[_a, _b, _c].T, cmap=cmap,
                                origin='lower',
                                extent=extent_val, interpolation=interpolation)
                if legend:
                    helpers.add_colorbar(im, label='property value', location='right')

        else:
            f, ax = plt.subplots(1, 2, sharex=True, sharey=True)
            ax[0].title.set_text('Estimated value')
            im1 = ax[0].imshow(a.reshape(self.domain.sol.grid.regular_grid.resolution)[_a, _b, _c].T, cmap=cmap,
                               origin='lower', interpolation=interpolation,
                               extent=self.domain.sol.grid.regular_grid.extent[[0, 1, 4, 5]])
            helpers.add_colorbar(im1, label='property value')
            ax[1].title.set_text('Variance')
            im2 = ax[1].imshow(b.reshape(self.domain.sol.grid.regular_grid.resolution)[_a, _b, _c].T, cmap=cmap,
                               origin='lower', interpolation=interpolation,
                               extent=self.domain.sol.grid.regular_grid.extent[[0, 1, 4, 5]])
            helpers.add_colorbar(im2, label='variance[]')
            plt.tight_layout()

# TODO: check with new ordianry kriging and nugget effect
def simple_kriging(a, b, prop, var_mod, inp_mean):
    '''
    Method for simple kriging calculation.
    Args:
        a (np.array): distance matrix containing all distances between target point and moving neighbourhood
        b (np.array): distance matrix containing all inter-point distances between locations in moving neighbourhood
        prop (np.array): array containing scalar property values of locations in moving neighbourhood
        var_mod: variogram model object
    Returns:
        result (float?): single scalar property value estimated for target location
        std_ok (float?): single scalar variance value for estimate at target location
    '''

    # empty matrix building
    shape = len(a)
    C = np.zeros((shape, shape))
    c = np.zeros((shape))
    w = np.zeros((shape))

    # Filling matrices with covariances based on calculated distances
    C[:shape, :shape] = var_mod.calculate_covariance(b) #? cov or semiv
    c[:shape] = var_mod.calculate_covariance(a) #? cov or semiv

    # nugget effect for simple kriging - dont remember why i set this actively, should be the same
    #np.fill_diagonal(C, self.sill)

    # TODO: find way to check quality of matrix and solutions for instability
    # Solve Kriging equations
    w = np.linalg.solve(C, c)

    # calculating estimate and variance for kriging
    pred_var = var_mod.sill - np.sum(w * c)
    # Note that here the input mean is required, if kriged mean equivalent to OK
    result = inp_mean + np.sum(w * (prop - inp_mean))

    return result, pred_var

def ordinary_kriging(a, b, prop, var_mod):
    '''
    Method for ordinary kriging calculation.
    Args:
        a (np.array): distance matrix containing all distances between target point and moving neighbourhood
        b (np.array): distance matrix containing all inter-point distances between locations in moving neighbourhood
        prop (np.array): array containing scalar property values of locations in moving neighbourhood
        var_mod: variogram model object
    Returns:
        result (float?): single scalar property value estimated for target location
        std_ok (float?): single scalar variance value for estimate at target location
    '''

    # empty matrix building for OK
    shape = len(a)
    C = np.zeros((shape + 1, shape + 1))
    c = np.zeros((shape + 1))
    w = np.zeros((shape + 1))

    # filling matirces based on model for spatial correlation
    C[:shape, :shape] = var_mod.calculate_semivariance(b)
    c[:shape] = var_mod.calculate_semivariance(a)

    # matrix setup - compare pykrige, special for OK
    np.fill_diagonal(C, 0)  # this needs to be done as semivariance for distance 0 is 0 by definition
    C[shape, :] = 1.0
    C[:, shape] = 1.0
    C[shape, shape] = 0.0
    c[shape] = 1.0

    # This is if we want exact interpolator
    # but be aware that it strictly forces estimates to go through data points
    # c[c == self.nugget] = 0

    # TODO: find way to check quality of matrix and solutions for instability
    # Solve Kriging equations
    w = np.linalg.solve(C, c)

    # calculating estimate and variance for kriging
    pred_var = w[shape] + np.sum(w[:shape] * c[:shape])
    result = np.sum(w[:shape] * prop)

    return result, pred_var

def create_kriged_field(domain, variogram_model, distance_type='euclidian',
                        moving_neighbourhood='all', kriging_type='OK', n_closest_points=20):
    '''
    Method to create a kriged field over the defined grid of the gempy solution depending on the defined
    input data (conditioning).
    Returns:
        self.results_df (pandas dataframe):   Dataframe containing coordinates, kriging estimate
                                                    and kriging variance for each grid point
    '''
    # empty arrays for results (estimated values and variances)
    kriging_result_vals = np.zeros(len(domain.krig_grid))
    kriging_result_vars = np.zeros(len(domain.krig_grid))

    # Start with distance calculation
    # 1) all grid points to all data points
    # 2) all data points among each other
    if distance_type == 'euclidian':
        # calculate distances between all input data points
        dist_all_to_all = cdist(domain.data[:, :3], domain.data[:, :3])
        # calculate distances between all grid points and all input data points
        dist_grid_to_all = cdist(domain.krig_grid, domain.data[:, :3])

    # Main loop that goes through whole domain (grid)
    for i in range(len(domain.krig_grid)):

        # STEP 1: Multiple if elif conditions to define moving neighbourhood:
        if moving_neighbourhood == 'all':
            # cutting matrices and properties based on moving neighbourhood
            a = dist_grid_to_all[i]
            b = dist_all_to_all
            prop = domain.data[:, 3]

        elif moving_neighbourhood == 'n_closest':
            # cutting matrices and properties based on moving neighbourhood
            a = np.sort(dist_grid_to_all[i])
            a = a[:n_closest_points]
            aux = np.argsort(dist_grid_to_all[i])
            prop = domain.data[:, 3][aux]
            prop = prop[:n_closest_points]
            aux = aux[:n_closest_points]
            b = dist_all_to_all[np.ix_(aux, aux)]

        elif moving_neighbourhood == 'range':
            # cutting matrices and properties based on moving neighbourhood
            aux = np.where(dist_grid_to_all[i] <= variogram_model.range_)[0]
            a = dist_grid_to_all[i][aux]
            prop = domain.data[:, 3][aux]
            b = dist_all_to_all[np.ix_(aux, aux)]

        else:
            print("FATAL ERROR: Moving neighbourhood not understood")

        # STEP 2: Multiple if elif conditions to calculate kriging at point
        if kriging_type == 'OK':
            val, var = ordinary_kriging(a, b, prop, variogram_model)
        elif kriging_type == 'SK':
            val, var = simple_kriging(a, b, prop, variogram_model, domain.inp_mean)
        elif kriging_type == 'UK':
            print("Universal Kriging not implemented")
        else:
            print("FATAL ERROR: Kriging type not understood")

        # STEP 3: Save results
        kriging_result_vals[i] = val
        kriging_result_vars[i] = var

    # create dataframe of results data for calling
    d = {'X': domain.krig_grid[:, 0], 'Y': domain.krig_grid[:, 1], 'Z': domain.krig_grid[:, 2],
        'estimated value': kriging_result_vals, 'estimation variance': kriging_result_vars}

    results_df = pd.DataFrame(data=d)

    return field_solution(domain, variogram_model, results_df, field_type='interpolation')

def create_gaussian_field(domain, variogram_model, distance_type='euclidian',
                        moving_neighbourhood='all', kriging_type='OK', n_closest_points=20):
    '''
    Method to create a kriged field over the defined grid of the gempy solution depending on the defined
    input data (conditioning).
    Returns:
        self.results_df (pandas dataframe):   Dataframe containing coordinates, kriging estimate
                                                        and kriging variance for each grid point
    '''
    # perform SGS with same options as kriging
    # TODO: set options for no starting points (Gaussian field) - mean and variance

    # set random path through all unknown locations
    shuffled_grid = domain.krig_grid
    np.random.shuffle(shuffled_grid)

    # append shuffled grid to input locations
    sgs_locations = np.vstack((domain.data[:,:3],shuffled_grid))
    # create array for input properties
    sgs_prop_updating = domain.data[:,3] # use this and then always stack new ant end

    # container for estimation variances
    estimation_var = np.zeros(len(shuffled_grid))

    # - distance calculation (stays the same)
    # 1) all points to all points in order of path
    # 2) known locations at beginning?
    if distance_type == 'euclidian':
        # calculate distances between all input data points
        dist_all_to_all = cdist(sgs_locations, sgs_locations)

    # set counter og active data (start=input data, grwoing by 1 newly calcualted point each run)
    active_data = len(sgs_prop_updating)

    # Main loop that goes through whole domain (grid)
    for i in range(len(domain.krig_grid)):
        # STEP 1: cut update distance matrix to correct size
        # HAVE TO CHECK IF THIS IS REALLY CORRECT
        active_distance_matrix = dist_all_to_all[:active_data,:active_data]
        active_distance_vector = dist_all_to_all[:,active_data] #basically next point to be simulated
        active_distance_vector = active_distance_vector[:active_data] #cut to left or diagonal

        # TODO: NEED PART FOR ZERO INPUT OR NO POINTS IN RANGE OR LESS THAN N POINTS

        # STEP 2: Multiple if elif conditions to define moving neighbourhood:
        if moving_neighbourhood == 'all':
            # cutting matrices and properties based on moving neighbourhood
            a = active_distance_vector
            b = active_distance_matrix
            prop = sgs_prop_updating

        elif moving_neighbourhood == 'n_closest':
            # cutting matrices and properties based on moving neighbourhood

            # This seems to work
            if len(sgs_prop_updating) <= n_closest_points:
                a = active_distance_vector[:active_data]
                b = active_distance_matrix[:active_data,:active_data]
                prop = sgs_prop_updating

            # this does not # DAMN THIS STILL HAS ITSELF RIGHT? PROBLEM!
            else:
                a = np.sort(active_distance_vector)
                a = a[:n_closest_points]
                aux = np.argsort(active_distance_vector)
                prop = sgs_prop_updating[aux]
                prop = prop[:n_closest_points]
                aux = aux[:n_closest_points]
                b = active_distance_matrix[np.ix_(aux, aux)]

        elif moving_neighbourhood == 'range':
            # cutting matrices and properties based on moving neighbourhood
            aux = np.where(active_distance_vector <= variogram_model.range_)[0]
            a = active_distance_vector[aux]
            prop = sgs_prop_updating[aux]
            b = active_distance_matrix[np.ix_(aux, aux)]

        else:
            print("FATAL ERROR: Moving neighbourhood not understood")

        # STEP 3: Multiple if elif conditions to calculate kriging at point
        # TODO: Cover case of data location and grid point coinciding
        if kriging_type == 'OK':
            val, var = ordinary_kriging(a, b, prop, variogram_model)
        elif kriging_type == 'SK':
            val, var = simple_kriging(a, b, prop, variogram_model, domain.inp_mean)
        elif kriging_type == 'UK':
            print("Universal Kriging not implemented")
        else:
            print("FATAL ERROR: Kriging type not understood")

        # STEP 4: Draw from random distribution
        std_ = np.sqrt(var)
        estimate = np.random.normal(val, scale=std_)

        # append to prop:
        sgs_prop_updating = np.append(sgs_prop_updating, estimate)
        estimation_var[i]= var

        # at end of loop: include simulated point for next step
        active_data += 1

    # delete original input data from results
    simulated_prop = sgs_prop_updating[len(domain.data[:,3]):] # check if this works like intended

    # create dataframe of results data for calling
    d = {'X': shuffled_grid[:, 0], 'Y': shuffled_grid[:, 1], 'Z': shuffled_grid[:, 2],
         'estimated value': simulated_prop, 'estimation variance': estimation_var}

    results_df = pd.DataFrame(data=d)
    results_df = results_df.sort_values(['X','Y','Z'])

    return field_solution(domain, variogram_model, results_df, field_type='simulation')






import gempy_legacy as gp
import pyvista as pv
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import warnings

try:
    import faulthandler
    faulthandler.enable()
except Exception as e:  # pragma: no cover
    warnings.warn('Unable to enable faulthandler:\n%s' % str(e))


def test_pyvista_running():
    from pyvista import demos
    demos.plot_wave()

def test_plot_3d_data_default(one_fault_model_no_interp):
    gpv = gp.plot.plot_3d(one_fault_model_no_interp,
                          plotter_type='basic', off_screen=True, image=True, notebook=False)
    # img = gpv.p.show(screenshot=True)
    # plt.imshow(img[1])
    # plt.show()


def test_plot_3d_geo_map(unconformity_model):
    gpv = gp.plot.plot_3d(unconformity_model,
                          plotter_type='basic', off_screen=True,
                          show_topography=True,
                          show_scalar=False,
                          show_lith=False,
                          image=True,
                          kwargs_plot_structured_grid={'opacity': .5})
    # img = gpv.p.show(screenshot=True)
    # plt.imshow(img[1])
    # plt.show()


def test_plot_3d_geo_map2(one_fault_model_topo_solution):
    gpv = gp.plot.plot_3d(one_fault_model_topo_solution,
                          plotter_type='basic', off_screen=True,
                          show_topography=True,
                          show_scalar=False,
                          show_lith=False,
                          image=True,
                          kwargs_plot_structured_grid={'opacity': .5})
    # img = gpv.p.show(screenshot=True)
    # plt.imshow(img[1])
    # plt.show()


def test_plot_3d_structure_topo(one_fault_model_topo_solution):
    one_fault_model_topo_solution._grid.regular_grid.set_topography_mask(
        one_fault_model_topo_solution._grid.topography)
    gpv = gp.plot.plot_3d(one_fault_model_topo_solution,
                          plotter_type='basic', off_screen=True,
                          show_topography=True,
                          show_scalar=False,
                          show_lith=True,
                          image=True,
                          kwargs_plot_structured_grid={'opacity': .5})
    # img = gpv.p.show(screenshot=True)
    # plt.imshow(img[1])
    # plt.show()


def test_plot_3d_structure_topo2(unconformity_model_topo):
    gpv = gp.plot.plot_3d(unconformity_model_topo,
                          plotter_type='basic', off_screen=True,
                          show_topography=True,
                          show_scalar=False,
                          show_lith=True,
                          image=True,
                          kwargs_plot_structured_grid={'opacity': .5})
    # img = gpv.p.show(screenshot=True)
    # plt.imshow(img[1])
    # plt.show()



"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import aesara
import aesara.tensor  as T
from gempy_legacy.core.grid_modules.grid_types import CenteredGrid


class GravityPreprocessing(CenteredGrid):
    def __init__(self, centered_grid: CenteredGrid = None):

        if centered_grid is None:
            super().__init__()
        elif isinstance(centered_grid, CenteredGrid):
            self.kernel_centers = centered_grid.kernel_centers
            self.kernel_dxyz_right = centered_grid.kernel_dxyz_right
            self.kernel_dxyz_left = centered_grid.kernel_dxyz_left
        self.tz = np.empty(0)

    def set_tz_kernel(self, scale=True, **kwargs):
        if self.kernel_centers.size == 0:
            self.set_centered_kernel(**kwargs)

        grid_values = self.kernel_centers

        s_gr_x = grid_values[:, 0]
        s_gr_y = grid_values[:, 1]
        s_gr_z = grid_values[:, 2]

        # getting the coordinates of the corners of the voxel...
        x_cor = np.stack((s_gr_x - self.kernel_dxyz_left[:, 0], s_gr_x + self.kernel_dxyz_right[:, 0]), axis=1)
        y_cor = np.stack((s_gr_y - self.kernel_dxyz_left[:, 1], s_gr_y + self.kernel_dxyz_right[:, 1]), axis=1)
        z_cor = np.stack((s_gr_z - self.kernel_dxyz_left[:, 2], s_gr_z + self.kernel_dxyz_right[:, 2]), axis=1)

        # ...and prepare them for a vectorial op
        x_matrix = np.repeat(x_cor, 4, axis=1)
        y_matrix = np.tile(np.repeat(y_cor, 2, axis=1), (1, 2))
        z_matrix = np.tile(z_cor, (1, 4))

        s_r = np.sqrt(x_matrix ** 2 + y_matrix ** 2 + z_matrix ** 2)

        # This is the vector that determines the sign of the corner of the voxel
        mu = np.array([1, -1, -1, 1, -1, 1, 1, -1])

        if scale is True:
            #
            G = 6.674e-3 # ugal     cm3⋅g−1⋅s−26.67408e-2 -- 1 m/s^2 to milligal = 100000 milligal
        else:
            from scipy.constants import G

        self.tz = (
            G *
            np.sum(- 1 *
                   mu * (
                           x_matrix * np.log(y_matrix + s_r) +
                           y_matrix * np.log(x_matrix + s_r) -
                           z_matrix * np.arctan(x_matrix * y_matrix / (z_matrix * s_r))),
                   axis=1))

        return self.tz


class MagneticsPreprocessing(CenteredGrid):
    """
    @Nilgün Güdük

    """
    def __init__(self, centered_grid: CenteredGrid = None):

        if centered_grid is None:
            super().__init__()
        elif isinstance(centered_grid, CenteredGrid):
            self.kernel_centers = centered_grid.kernel_centers
            self.kernel_dxyz_right = centered_grid.kernel_dxyz_right
            self.kernel_dxyz_left = centered_grid.kernel_dxyz_left
        self.V = np.empty(0)

    def set_Vs_kernel(self, **kwargs):
        if self.kernel_centers.size == 0:
            self.set_centered_kernel(**kwargs)

        grid_values = self.kernel_centers
        s_gr_x = grid_values[:, 0]
        s_gr_y = grid_values[:, 1]
        s_gr_z = -1 * grid_values[:, 2]  # talwani takes x-axis positive downwards, and gempy negative downwards

        # getting the coordinates of the corners of the voxel...
        x_cor = np.stack((s_gr_x - self.kernel_dxyz_left[:, 0], s_gr_x + self.kernel_dxyz_right[:, 0]), axis=1)
        y_cor = np.stack((s_gr_y - self.kernel_dxyz_left[:, 1], s_gr_y + self.kernel_dxyz_right[:, 1]), axis=1)
        z_cor = np.stack((s_gr_z + self.kernel_dxyz_left[:, 2], s_gr_z - self.kernel_dxyz_right[:, 2]), axis=1)
        # ...and prepare them for a vectorial op
        x_matrix = np.repeat(x_cor, 4, axis=1)
        y_matrix = np.tile(np.repeat(y_cor, 2, axis=1), (1, 2))
        z_matrix = np.tile(z_cor, (1, 4))

        R = np.sqrt(x_matrix ** 2 + y_matrix ** 2 + z_matrix ** 2)  # distance to each corner
        s = np.array([-1, 1, 1, -1, 1, -1, -1, 1])  # gives the sign of each corner: depends on your coordinate system

        # variables V1-6 represent integrals of volume for each voxel
        V1 = np.sum(-1 * s * np.arctan2((y_matrix * z_matrix), (x_matrix * R)), axis=1)
        V2 = np.sum(s * np.log(R + z_matrix), axis=1)
        V3 = np.sum(s * np.log(R + y_matrix), axis=1)
        V4 = np.sum(-1 * s * np.arctan2((x_matrix * z_matrix), (y_matrix * R)), axis=1)
        V5 = np.sum(s * np.log(R + x_matrix), axis=1)
        V6 = np.sum(-1 * s * np.arctan2((x_matrix * y_matrix), (z_matrix * R)), axis=1)

        # contains all the volume integrals (6 x n_kernelvalues)
        V = np.array([V1, V2, V3, V4, V5, V6])
        return V



"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


    @author: Elisa Heim, Alexander Schaaf, Miguel de la Varga
    (I guess, copied some code from posterior_analysis_DEP.py)
"""

import warnings
try:
    import pymc
except ImportError:
    warnings.warn("pymc (v2) package is not installed. No support for stochastic simulation posterior analysis.")
import numpy as np
import pandas as pn
import gempy_legacy as gp
try:
    import tqdm
except ImportError:
    warnings.warn("tqdm package not installed. No support for dynamic progress bars.")
import matplotlib.pyplot as plt
from mpl_toolkits import axes_grid1
import matplotlib.colors


class Posterior():
    def __init__(self, dbname, model_type='map', entropy=False,
                 topography=None, interpdata=None, geodata=None):

        if entropy:
            print('All post models are calculated. Based on the model complexity and the number of iterations, '
                  'this could take a while')
        # if topography:
        self.topography = topography
        # else:
        # print('no topography defined. Methods that contain the word _map_ are not available')

        self.interp_data = interpdata

        self.geo_data = geodata
        # self.verbose = verbose

        self.db = pymc.database.hdf5.load(dbname)  # load database
        self.n_iter = self.db.getstate()['sampler']['_iter'] - self.db.getstate()["sampler"]["_burn"]
        self.trace_names = self.db.trace_names[0]
        self.input_data = self.db.input_data.gettrace()

        if entropy:
            if topography and model_type == 'map':  # better resolution
                self.all_maps = self.all_post_maps()
                self.map_prob = self.compute_prob(np.round(self.all_maps).astype(int))
                self.map_ie = self.calculate_ie_masked(self.map_prob)

            elif model_type == 'model':
                self.lbs, self.fbs = self.all_post_models()

                if len(self.lbs) != 0:
                    self.lith_prob = self.compute_prob(np.round(self.lbs).astype(int))
                    self.lb_ie = self.calculate_ie_masked(self.lith_prob)

                if len(self.fbs) != 0:
                    self.fault_prob = self.compute_prob(np.round(self.fbs).astype(int))
                    self.fb_ie = self.calculate_ie_masked(self.fault_prob)
            else:
                print('if there is no topography defined, model_type must be set to model')
            # self.ie_total = self.calculate_ie_total()

    def _change_input_data(self, i):
        i = int(i)
        # replace interface data
        self.interp_data.geo_data_res.interfaces[["X", "Y", "Z"]] = self.input_data[i][0]
        # replace foliation data
        self.interp_data.geo_data_res._orientations[["G_x", "G_y", "G_z", "X", "Y", "Z", "dip", "azimuth", "polarity"]] = \
            self.input_data[i][1]
        self.interp_data.update_interpolator()
        # if self.verbose:
        # print("interp_data parameters changed.")
        return self.interp_data

    def all_post_maps(self):
        all_maps = []
        for i in range(0, self.n_iter):
            # print(i)
            self._change_input_data(i)
            # geomap = self.topography.calculate_geomap(interpdata = self.interp_data, plot=True)
            geomap, faultmap = gp.compute_model_at(self.topography.surface_coordinates[0], self.interp_data)
            all_maps.insert(i, geomap[0])
        return all_maps

    def all_post_models(self):
        lbs = []
        fbs = []
        for i in range(0, self.n_iter):
            # print(i)
            self._change_input_data(i)
            lith_block, fault_block = gp.compute_model(self.interp_data)
            if lith_block.shape[0] != 0:
                lbs.insert(i, lith_block[0])
            if fault_block.shape[0] != 0:
                n = 0
                while n < fault_block.shape[0]:
                    # print(fault_block.shape[0])
                    fbs.insert(i, fault_block[n])
                    n += 2
        return lbs, fbs

    def compute_prob(self, blocks):
        lith_id = np.unique(blocks)
        # lith_count = np.zeros_like(lith_blocks[0:len(lith_id)])
        count = np.zeros((len(np.unique(blocks)), blocks.shape[1]))
        for i, l_id in enumerate(lith_id):
            count[i] = np.sum(blocks == l_id, axis=0)
        prob = count / len(blocks)
        # print(lith_prob)
        return prob

    def calculate_ie_masked(self, prob):
        ie = np.zeros_like(prob[0])
        for l in prob:
            pm = np.ma.masked_equal(l, 0)  # mask where prob is 0
            ie -= (pm * np.ma.log2(pm)).filled(0)
        return ie

    def calculate_ie_total(self, ie, absolute=False):
        if absolute:
            return np.sum(ie)
        else:
            return np.sum(ie) / np.size(ie)

        ##### plotting methods #####

    def plot_section(self, iteration=1, block='lith', cell_number=3, **kwargs):
        '''kwargs: gempy.plotting.plot_section keyword arguments'''
        self._change_input_data(iteration)
        lith_block, fault_block = gp.compute_model(self.interp_data)

        if 'topography' not in kwargs:
            if self.topography:
                topography = self.topography
            else:
                topography = None
            if block == 'lith':
                gp.plot_section(self.geo_data, lith_block[0], cell_number=cell_number, topography=topography, **kwargs)
            else:
                gp.plot_section(self.geo_data, block, cell_number=cell_number, topography=topography, **kwargs)

        else:
            if block == 'lith':
                gp.plot_section(self.geo_data, lith_block[0], cell_number=cell_number, **kwargs)
            else:
                gp.plot_section(self.geo_data, block, cell_number=cell_number, **kwargs)

    def plot_map(self, iteration=1, **kwargs):
        self._change_input_data(iteration)
        # geomap = self.topography.calculate_geomap(interpdata = self.interp_data, plot=True)
        geomap, faultmap = gp.compute_model_at(self.topography.surface_coordinates[0], self.interp_data)
        # gp.plotting.plot_map(geomap)
        gp.plotting.plot_map(self.geo_data, geomap=geomap[0].reshape(self.topography.dem_zval.shape), **kwargs)

    def plot_map_ie(self, plot_data=False):
        if plot_data:
            gp.plotting.plot_data(geo_data, direction='z')
            dist = 12
        else:
            dist = 1

        im = plt.imshow(self.map_ie.reshape(self.topography.dem_zval.shape), extent=self.geo_data.extent[:4],
                        cmap='viridis')
        self.add_colorbar(im, pad_fraction=dist)
        plt.title('Cell entropy of geological map')

    def plot_section_ie(self, block='lith', cell_number=10, direction='y', **kwargs):
        # for lithblock
        if block == 'lith':
            norm = matplotlib.colors.Normalize(self.lb_ie.min(), self.lb_ie.max())
            gp.plotting.plot_section(geo_data, self.lb_ie, cell_number=cell_number, direction=direction, cmap='viridis',
                                     norm=norm, **kwargs)
            # self.add_colorbar(im)
        elif block == 'fault':
            norm = matplotlib.colors.Normalize(self.fb_ie.min(), self.fb_ie.max())
            gp.plotting.plot_section(geo_data, self.fb_ie, cell_number=cell_number, direction=direction, cmap='viridis',
                                     norm=norm, **kwargs)
            # self.add_colorbar(im)

    def add_colorbar(self, im, aspect=20, pad_fraction=1, **kwargs):
        """Add a vertical color bar to an image plot. Source: stackoverflow"""
        divider = axes_grid1.make_axes_locatable(im.axes)
        width = axes_grid1.axes_size.AxesY(im.axes, aspect=2. / aspect)
        pad = axes_grid1.axes_size.Fraction(pad_fraction, width)
        current_ax = plt.gca()
        cax = divider.append_axes("right", size=width, pad=pad)
        plt.sca(current_ax)
        return im.axes.figure.colorbar(im, cax=cax, **kwargs)



import os
import gempy_legacy as gp
from gempy_legacy.bayesian.aesara_op import GemPyThOp
import pytest
pm = pytest.importorskip("rgeomod")
# import pymc as pm
import arviz as az
import numpy as np
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import aesara
np.random.seed(4003)


def test_basic():
    y_obs_list = [2.12, 2.06, 2.08, 2.05, 2.08, 2.09,
                  2.19, 2.07, 2.16, 2.11, 2.13, 1.92]

    with pm.Model() as model:
        mu = pm.Normal('$\mu$', 2.08, .07)
        sigma = pm.Gamma('$\sigma$', 0.3, 3)
        y = pm.Normal('$y$', mu, sigma, observed=y_obs_list)

        prior = pm.sample_prior_predictive(1000)
        trace = pm.sample(1000, discard_tuned_samples=False, cores=1)
        post = pm.sample_posterior_predictive(trace)

    az.plot_posterior(trace)
    plt.show()


def test_gempy_th_op_test():
    path_dir = os.getcwd() + '/../../examples/tutorials/ch5_probabilistic_modeling'
    geo_model = gp.load_model(r'2-layers', path=path_dir, recompile=False)
    gto = GemPyThOp(geo_model)
    sol = gto.test_gradient('lith', 'surface_points')
    print(sol)


def test_gempy_th_op_set_grav():
    path_dir = os.getcwd() + '/../../examples/tutorials/ch5_probabilistic_modeling'
    geo_model = gp.load_model(r'2-layers', path=path_dir, recompile=False)
    gp.set_interpolator(geo_model, output='grav')

    gto = GemPyThOp(geo_model)
    th_op_grav = gto.set_th_op('gravity')
    i = geo_model.interpolator.get_python_input_block()
    th_f = aesara.function([], th_op_grav(*i), on_unused_input='warn')


import pytest

import aesara
import numpy as np
import sys, os

sys.path.append("../..")
import gempy_legacy
import matplotlib.pyplot as plt
import pdb


@pytest.mark.skip()
class TestPerformance:
    @pytest.mark.skip()
    def test_high_res(self):
        pass



"""Implementation of Diamond-Square algorithm

This algorithm is often used for random topography generation, following Fournier et al., 1982
see https://en.wikipedia.org/wiki/Diamond-square_algorithm.

Fournier, Alain; Fussell, Don; Carpenter, Loren (June 1982). "Computer rendering of stochastic models".
Communications of the ACM. 25 (6): 371–384.

Here the description from the wikipedia page:

+++ begin Wikipedia +++

The diamond-square algorithm begins with a 2D square array of width and
height 2n + 1. The four corner points of the array must first be set to initial values. The diamond and square steps
are then performed alternately until all array values have been set.

The diamond step: For each square in the array, set the midpoint of that square to be the average of the four corner
points plus a random value.

The square step: For each diamond in the array, set the midpoint of that diamond to be the average of the four corner
points plus a random value.

At each iteration, the magnitude of the random value should be reduced.

During the square steps, points located on the edges of the array will have only three adjacent values set rather
than four. There are a number of ways to handle this complication - the simplest being to take the average of just
the three adjacent values. Another option is to 'wrap around', taking the fourth value from the other side of the
array. When used with consistent initial corner values this method also allows generated fractals to be stitched
together without discontinuities.

+++ end Wikipedia +++

The implementations (in Python) I could find online were either difficult to understand or had many case
selections, especially at edges. Here is a fully vectorized implementation, using padding at edges, to avoid
all these case selections for a more straight-forward implementation (I hope).

This implementation is also adjusted to work on non-square start matrices, and on reduced hierarchies with
more initial (internal) points.

Created on 10.04.2020

@author: Florian Wellmann

"""
import numpy as np
import matplotlib.pyplot as plt


class DiaomondSquare(object):

    def __init__(self, size: tuple = (16, 16), roughness: float = 0.5, z_min: float = 0, z_max:
                 float = 1, r_type: chr = 'default', **kwds):
        """Implementation of vectorized Diaomnd-Square algorithm for random topography generation

        Args:
            size (int, int): shape of grid to interpolate; note: the standard diamond-square algorithm
                operates on a square grid with side length 2**n+1. This implementation is adjusted to non-square
                grids with (2**n+1, 2**m+1). If the input size (int, int) is not matching to the ideal dimension,
                the next bigger size is taken and the grid finally cut (lower left corner is kept);
            roughness: roughness parameter, [0,1]: 0: deterministic interpolation, 1: very rough and bumpy (in standard
                randomisation method - other options available in self.random_func())
            z_min: minimum height of surface
            z_max: maximum height of surface
            seed: seed for random function to enable reproducibility and testing
            r_type: 'basic', 'level_scale' (See below for details)
            
        Options for the randomization function (self.r_type) are:
        - 'default': standard reduction
        - 'level_scale' : uniform distribution between z_min, z_max, scaled by (level + 1)

        """
        self.size = size
        # Create mesh with optimal size (2**n+1, 2**m+1)
        # If the input size `self.size` does not match to these dimensions, then the next larger suitable
        # size is chosen.
        self.n = np.ceil(np.log2(self.size[0] - 1)).astype('int8')
        self.m = np.ceil(np.log2(self.size[1] - 1)).astype('int8')
        self.grid = np.zeros((2 ** self.n + 1, 2 ** self.m + 1))
        self.roughness = roughness
        self.z_min = z_min
        self.z_max = z_max
        if 'seed' in kwds:
            np.random.seed(kwds['seed'])
        self.base_rand_range = 1
        self.r_type = r_type

    def random_func(self, i, level_shape):
        """Define random function to adjust range for subsequent steps

        The random function type is defined on class level (self.r_type)
        """

        if self.r_type == 'default':
            if self.roughness > 0:
                rand_range = self.base_rand_range * 2 ** (- (1 - self.roughness) * i)
                return np.random.uniform(-rand_range, rand_range, level_shape)
            else:
                return 0.

        if self.r_type == 'long_range':
            if self.roughness > 0:
                rand_range = self.base_rand_range * 16 ** (- (1 - self.roughness) * i)
                return np.random.uniform(-rand_range, rand_range, level_shape)
            else:
                return 0.

        elif self.r_type == 'level_scale':
            if self.roughness > 0:
                return np.random.uniform(self.z_min, self.z_max,
                                         level_shape) / (1 + i)
            else:
                return 0.

        elif self.r_type == 'deterministic':
            # No randomization at all - deterministic interpolation result
            return 0.

        else:
            raise NotImplementedError("Random function type %s not implemented" % self.r_type)

    def random_initialization(self, level='highest', verbose: bool=False):
        """Initialize cells on speicifc hierarchy with random values

        Args:
            level = 'hightest', int : hierarchy level for interpolation (default: highest)
            verbose: verbose output (default: False)

        With highest hierarchy, we refer here to the largest diamond-square step, i.e. the corner points
        for a square grid; Or, more formally: the diamond points for `min(self.n, self.m)`
        """
        if level == 'highest':
            m_pow_max = min(self.n, self.m)
        else:
            m_pow_max = level

        step_size = int(2 ** m_pow_max)
        if verbose:
            print("Initialize on step size %d" % step_size)

        level_shape = self.grid[::step_size, ::step_size].shape

        self.grid[::step_size, ::step_size] = np.random.uniform(0, 1, level_shape)  # self.random_func(0, level_shape)

    def interpolate(self, level='highest'):
        """Perform diamond-square interpolation

        Args:
            level = 'hightest', int : hierarchy level for interpolation (default: highest)

        This step follows the conventional procedure:
        Iterate over hierarchies and repeat:
        2) Perform Diamond interpolation step
        3) Perform Square interpolation step
        4) Reduce roughness factor
        """
        if level == 'highest':
            # determine highest hierarchy level (determined by shorter rectangle side)
            m_pow_max = min(self.n, self.m)
        else:
            m_pow_max = level

        for i, m_pow in enumerate(np.arange(m_pow_max)[::-1]):
            self.perform_diamond_step(i, m_pow)
            self.perform_square_step(i, m_pow)

    def reset_grid(self):
        """Reset grid back to zero values"""
        self.grid[:, :] = 0

    def perform_diamond_step(self, i: int, m_pow: int):
        """Perform one diamond interpolation step on hierarchy m_pow

        Note: for more details on the vectorized selection, see self.get_selection_diamond()
        """
        step_size = int(2 ** m_pow)

        # Diamond step
        # ----------------

        # get shape of this step
        level_shape = self.grid[step_size::2 * step_size, step_size::2 * step_size].shape

        self.grid[step_size::2 * step_size, step_size::2 * step_size] = \
            (self.grid[:-2 * step_size:2 * step_size, :-2 * step_size:2 * step_size] +
             self.grid[:-2 * step_size:2 * step_size, 2 * step_size::2 * step_size] +
             self.grid[2 * step_size::2 * step_size, :-2 * step_size:2 * step_size] +
             self.grid[2 * step_size::2 * step_size, 2 * step_size::2 * step_size]) / \
            4. + \
            self.random_func(i, level_shape)

        # np.random.uniform(-rand_range, rand_range, level_shape)

        # self.d * 2**()
        # self.roughness ** i * (np.random.random(step_shape) - 0.5)

        # (np.random.random(step_shape) - 0.5) * i * self.roughness

    def perform_square_step(self, i: int, m_pow: int):
        """Perform one square interpolation step on hierarchy m_pow

        Note: for more details on the vectorized selection, see self.get_selection_square()
        """
        step_size = int(2 ** m_pow)

        # pad cells with zero value
        z_pad = np.pad(self.grid, step_size, mode='constant')

        # also create a grid for division to divide only by 3 on borders
        grid_div = np.ones_like(self.grid[1:-1, 1:-1]) * 4.
        grid_div = np.pad(grid_div, step_size + 1, mode='constant', constant_values=3.)

        # Checkerboard odd
        # ----------------

        # get shape of this step
        level_shape = grid_div[step_size::2 * step_size, 2 * step_size:-2 * step_size:2 * step_size].shape

        z_pad[step_size::2 * step_size, 2 * step_size:-2 * step_size:2 * step_size] = \
            (z_pad[step_size::2 * step_size, step_size:-2 * step_size:2 * step_size] +
             z_pad[step_size::2 * step_size, 3 * step_size:-step_size:2 * step_size] +
             z_pad[:-step_size:2 * step_size, 2 * step_size:-2 * step_size:2 * step_size] +
             z_pad[2 * step_size::2 * step_size, 2 * step_size:-2 * step_size:2 * step_size]) / \
            grid_div[step_size::2 * step_size, 2 * step_size:-2 * step_size:2 * step_size] + \
            self.random_func(i, level_shape)

        # np.random.uniform(-rand_range, rand_range, level_shape)

        #    self.roughness ** i * (np.random.random(step_shape) - 0.5)

        # (np.random.random(step_shape) - 0.5) * i * self.roughness

        # Checkerboard even
        # -----------------

        # get shape of this step
        level_shape = z_pad[2 * step_size:-2 * step_size:2 * step_size, step_size:-step_size:2 * step_size].shape

        # check-even, values to interpolate:
        z_pad[2 * step_size:-2 * step_size:2 * step_size, step_size:-step_size:2 * step_size] = \
            (z_pad[2 * step_size:-2 * step_size:2 * step_size, :-2 * step_size:2 * step_size] +
             z_pad[2 * step_size:-2 * step_size:2 * step_size, 2 * step_size::2 * step_size] +
             z_pad[step_size:-2 * step_size:2 * step_size, step_size:-step_size:2 * step_size] +
             z_pad[3 * step_size::2 * step_size, step_size:-step_size:2 * step_size]) / \
            grid_div[2 * step_size:-2 * step_size:2 * step_size, step_size:-step_size:2 * step_size] + \
            self.random_func(i, level_shape)

        # np.random.uniform(-rand_range, rand_range, level_shape)

        # self.roughness ** i * (np.random.random(step_shape) - 0.5)

        # (np.random.random(step_shape) - 0.5) * i * self.roughness

        # assign results back to self.grid
        self.grid = z_pad[step_size:-step_size, step_size:-step_size]

    def get_selection_diamond(self, m_pow: int):
        """get selected points for diamond step on grid z on hierarchy m

        This method is mostly implemented for testing and visualization purposes.
        """

        step_size = int(2 ** m_pow)

        z = np.zeros_like(self.grid, dtype='int8')

        # points to interpolate
        z[step_size::2 * step_size, step_size::2 * step_size] = 1

        # top left
        z[:-2 * step_size:2 * step_size, :-2 * step_size:2 * step_size] = 2

        # top right
        z[:-2 * step_size:2 * step_size, 2 * step_size::2 * step_size] = 2

        # bottom left
        z[2 * step_size::2 * step_size, :-2 * step_size:2 * step_size] = 2

        # bottom right
        z[2 * step_size::2 * step_size, 2 * step_size::2 * step_size] = 2

        return z

    def get_selection_square(self, m_pow: int):
        """Plot selected points for square step on grid z on hierarchy m

        This method is mostly implemented for testing and visualization purposes.
        """
        z = np.zeros_like(self.grid, dtype='int8')
        step_size = int(2 ** m_pow)

        # pad cells with zero value
        z_pad = np.pad(z, step_size, mode='constant')

        # Checkerboard odd
        # ----------------

        # check-odd, values to interpolate:
        z_pad[step_size::2 * step_size, 2 * step_size:-2 * step_size:2 * step_size] = 1

        # check-odd, left
        z_pad[step_size::2 * step_size, step_size:-2 * step_size:2 * step_size] = 2

        # check-odd, right
        z_pad[step_size::2 * step_size, 3 * step_size:-step_size:2 * step_size] = 2

        # check-odd, top
        z_pad[:-step_size:2 * step_size, 2 * step_size:-2 * step_size:2 * step_size] = 2

        # check-odd, bottom
        z_pad[2 * step_size::2 * step_size, 2 * step_size:-2 * step_size:2 * step_size] = 2

        # Checkerboard even
        # -----------------

        # check-even, values to interpolate:
        z_pad[2 * step_size:-2 * step_size:2 * step_size, step_size:-step_size:2 * step_size] = 1

        # check-even, left:
        z_pad[2 * step_size:-2 * step_size:2 * step_size, :-2 * step_size:2 * step_size] = 2

        # check-even, right:
        z_pad[2 * step_size:-2 * step_size:2 * step_size, 2 * step_size::2 * step_size] = 2

        # check-even, top:
        z_pad[step_size:-2 * step_size:2 * step_size, step_size:-step_size:2 * step_size] = 2

        # check-even, bottom:
        z_pad[3 * step_size::2 * step_size, step_size:-step_size:2 * step_size] = 2

        return z_pad

    def plot_diamond_and_square(self, pad=False):
        """Plot selected points for diamond and square step for all hierarchies side by side"""

        m_pow_max = min(self.n, self.m)

        shape_ratio = self.n / self.m

        f, axes = plt.subplots(2, m_pow_max, figsize=(12, 12 * shape_ratio / m_pow_max * 2))

        for i, m_pow in enumerate(np.arange(m_pow_max)[::-1]):
            m = 2 ** m_pow
            # z_zero = np.zeros_like(self.grid)
            z_diamond = self.get_selection_diamond(m_pow)
            z_square = self.get_selection_square(m_pow)
            if pad:
                z_pad = np.pad(z_diamond, m, mode='constant')
                axes[0, i].imshow(z_pad, cmap='viridis', vmin=0, vmax=2)
                axes[1, i].imshow(z_square, cmap='viridis', vmin=0, vmax=2)
            else:
                axes[0, i].imshow(z_diamond, cmap='viridis', vmin=0, vmax=2)
                axes[1, i].imshow(z_square[m:-m, m:-m], cmap='viridis', vmin=0, vmax=2)



import pytest
from gempy_legacy.assets import topology as tp
import numpy as np
import gempy_legacy as gp
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt


@pytest.fixture(scope='module')
def topology_fabian(one_fault_model_solution):
    """Return a GemPy Vista instance with basic geomodel attached."""
    geo_model = one_fault_model_solution

    edges, centroids = tp.compute_topology(geo_model, voxel_threshold=1)

    return edges, centroids


@pytest.fixture(scope='module')
def topology_jan_unconf(unconformity_model):
    geo_model = unconformity_model
    gp.plot.plot_2d(unconformity_model)
    plt.show()
    edges, centroids = tp.compute_topology(geo_model, voxel_threshold=1)
    return edges, centroids


def assert_centroids(centroids, centroids_test):
    for key in centroids_test.keys():
        for i in range(2):
            assert pytest.approx(centroids[key][i]) == centroids_test[key][i]


@pytest.fixture
def edges_fabian_test():
    return {(1, 2), (4, 10), (2, 7), (6, 7), (5, 10), (4, 9), (4, 5), (2, 8),
            (3, 8), (8, 9), (9, 10), (3, 9), (2, 3), (1, 6), (1, 7), (3, 4),
            (7, 8)}


@pytest.fixture
def centroids_fabian_test():
    return {1: np.array([12.16046912, 24.8310141, 40.65527994]),
            2: np.array([8.93961983, 24.92955647, 28.73667536]),
            3: np.array([7.95101873, 24.71784318, 23.43445153]),
            4: np.array([6.85111073, 24.53773105, 17.19230117]),
            5: np.array([4.92739115, 23.20962489, 7.25340731]),
            6: np.array([35.9145365, 25.00458115, 39.16981059]),
            7: np.array([33.7571116, 24.83114515, 26.98018478]),
            8: np.array([32.89434051, 24.74326391, 21.65431227]),
            9: np.array([32.08333333, 24.65390105, 15.5167301]),
            10: np.array([31.04821962, 23.66595199, 5.89393366])}


def test_edges_fabian(topology_fabian, edges_fabian_test):
    assert topology_fabian[0] == edges_fabian_test


def test_centroids_fabian(topology_fabian, centroids_fabian_test):
    assert_centroids(topology_fabian[1], centroids_fabian_test)


@pytest.fixture
def edges_jan_unconf_test():
    return {(1, 2), (1, 3), (1, 4), (2, 3), (3, 4)}


@pytest.fixture
def centroids_jan_unconf_test():
    return {1: np.array([24.5, 20.5, 27.5]),
            2: np.array([24.5, 20.5, 21.37869822]),
            3: np.array([24.5, 20.5, 19.12100139]),
            4: np.array([24.5, 20.5, 9.71685136])}


def test_edges_jan_unconf(topology_jan_unconf, edges_jan_unconf_test):
    assert topology_jan_unconf[0] == edges_jan_unconf_test


def test_centroids_jan_unconf(topology_jan_unconf, centroids_jan_unconf_test):
    assert_centroids(topology_jan_unconf[1], centroids_jan_unconf_test)



"""Class definition for GeoModeller XML-Files
This version includes drillholes
Specific methods are defined for the uncertainty analysis (in combination
with Uncertainty_Obj module)

(c) J. Florian Wellmann, 2009-2013
"""

# try:
#     import elementtree.ElementTree as ET
# except ImportError:
#     try:
#         import etree.ElementTree as ET
#     except ImportError:
#         try:
#             import xml.etree.ElementTree as ET
#         except ImportError:
#             import ElementTree as ET
# import Latex_output_5 as LO
from pylab import *
import copy
import pandas as pn
import gempy_legacy as gp
import numpy as np

# python module to wrap GeoModeller XML file and perform all kinds of data
# procedures, e.g.:
# - Stochastic data modeling
# - TWT to depth conversion
# - Documentation module
#
# To Do:
# - exception handling and passing
# - log file?
# - check validity of xml-code, test, if past-processing with sax neccessary?
# - implement auto-documentation
# - clear-up spaghetti code!!!!! Check dependencies and other modules
#    to get a consistent lay-out

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET

# TODO fix this function and importing it to gempy.utils.create_from_geomodeller_xml
def create_from_geomodeller_xml(fp, resolution=(50, 50, 50), return_xml=False, **kwargs):
    """
    EXPERIMENTAL
    Creates InputData object from a GeoModeller xml file. Automatically extracts and sets model extent, interface
    and orientation data as well as the stratigraphic pile.

    Args:
        fp (str): Filepath for the GeoModeller xml file to be read.
        resolution (tuple, optional): Tuple containing the number of voxels in dimensions (x,y,z). Defaults to 50.
        return_xml (bool, optional): Toggles returning the ReadGeoModellerXML instance to leverage further info from the
            xml file (e.g. for stratigraphic pile ordering). Defaults to True.
        **kwargs: Keyword arguments for create_data function.

    Returns:
        gp.data_management.InputData
    """
    gmx = _ReadGeoModellerXML(fp)  # instantiate parser class with filepath of xml

    # instantiate InputData object with extent and resolution
    geo_data = create_data(gmx.extent, resolution, **kwargs)

    # set interface and orientation dataframes
    geo_data._surface_points = gmx._surface_points
    geo_data._orientations = gmx._orientations

    if return_xml:
        return geo_data, gmx
    else:
        return geo_data



class ReadGeoModellerXML:
    def __init__(self, fp):
        """
        Reads in and parses a GeoModeller XML file to extract interface and orientation data and the overall model
        settings (e.g. extent and sequential pile). It uses ElementTree to parse the XML and the tree's root can
        be accessed using self.root for more direct access to the file.

        Args:
            fp (str): Filepath for the GeoModeller xml file to be read.

        """
        self.tree = ET.ElementTree(file=fp)  # load xml as tree
        self.root = self.tree.getroot()

        self.xmlns = "http://www.geomodeller.com/geo"
        self.gml = "http://www.opengis.net/gml"

        self.extent = self._get_extent()

        self.data = self.extract_data()
        self.surface_points, self.orientations = self.get_dataframes()

        # self.stratigraphic_column = self.get_stratigraphic_column()
        # self.df = self.get_faults()
        #
        # self.series_info = self._get_series_fmt_dict()
        # self.series_distribution = self.get_series_distribution()
        #
        # self.fault_matrix = self.get_fault_matrix()

    def get_psc(self):
        """Returns the ProjectStratigraphicColumn tree element used for several data extractions."""
        return self.root.find("{" + self.xmlns + "}GeologicalModel").find("{"+self.xmlns+"}ProjectStratigraphicColumn")

    def extract_data(self):
        """
        Extracts relevant data from the GeoModeller XML file ElementTree root (self.root) and returns it as a dictionary.


        Returns:
            (dict): Data dictionary
        """
        data = {}
        for s in self.get_psc():
            sn = s.get("name")
            data[sn] = {}  # create a dict for each series
            data[sn]["formations"] = []
            data[sn]["InfluencedByFault"] = []
            data[sn]["relation"] = s.get("relation")  # add relation, whatever that is

            for c in s:
                if c.tag == "{"+self.xmlns+"}Data":  # append formation names to list of formations
                    data[sn]["formations"].append(c.get("Name"))

                if c.tag == "{"+self.xmlns+"}InfluencedByFault":  # add fault influences
                    data[sn]["InfluencedByFault"].append(c.get("Name"))

                if c.tag == "{"+self.xmlns+"}PotentialField":

                    data[sn]["gradients"] = []
                    data[sn]["surface_points"] = []
                    data[sn]["surface_points_counters"] = []
                    data[sn]["solutions"] = []
                    data[sn]["constraints"] = []

                    for cc in c:
                        # COVARIANCE
                        if cc.tag == "{" + self.xmlns + "}covariance":
                            data[sn]["covariance"] = cc.attrib

                        # GRADIENTS
                        if cc.tag == "{" + self.xmlns + "}Gradients":
                            for gr in cc:
                                data[sn]["gradients"].append([gr.get("Gx"), gr.get("Gy"), gr.get("Gz"),
                                                              gr.get("XGr"), gr.get("YGr"), gr.get("ZGr")])

                        # surface_points
                        if cc.tag == "{" + self.xmlns + "}Points":
                            for co in cc:
                                data[sn]["surface_points"].append([float(co[0].text), float(co[1].text), float(co[2].text)])

                        # INTERFACE COUNTERS
                        if cc.tag == "{" + self.xmlns + "}InterfacePoints":
                            for ip in cc:
                                data[sn]["surface_points_counters"].append([int(ip.get("npnt")), int(ip.get("pnt"))])

                        # CONSTRAINTS
                        if cc.tag == "{" + self.xmlns + "}Constraints":
                            for co in cc:
                                data[sn]["constraints"].append(float(co.get("value")))

                        # SOLUTIONS
                        if cc.tag == "{" + self.xmlns + "}Solutions":
                            for sol in cc:
                                data[sn]["solutions"].append(float(sol.get("sol")))

                    # convert from str to float
                    data[sn]["gradients"] = np.array(data[sn]["gradients"]).astype(float)
                    data[sn]["surface_points"] = np.array(data[sn]["surface_points"]).astype(float)
                    data[sn]["surface_points_counters"] = np.array(data[sn]["surface_points_counters"]).astype(float)
                    data[sn]["solutions"] = np.array(data[sn]["solutions"]).astype(float)

        return data

    def get_dataframes(self):
        """
        Extracts dataframe information from the self.data dictionary and returns GemPy-compatible surface_points and
        orientations dataframes.

        Returns:
            (tuple) of GemPy dataframes (surface_points, orientations)
        """
        interf_formation = []
        interf_series = []

        orient_series = []

        for i, s in enumerate(self.data.keys()):  # loop over all series
            if i == 0:
                coords = self.data[s]["surface_points"]
                grads = self.data[s]["gradients"]
            else:
                coords = np.append(coords, self.data[s]["surface_points"])
                grads = np.append(grads, self.data[s]["gradients"])

            for j, fmt in enumerate(self.data[s]["formations"]):
                for n in range(int(self.data[s]["surface_points_counters"][j, 0])):
                    interf_formation.append(fmt)
                    interf_series.append(s)

            for k in range(len(grads)):
                orient_series.append(s)

        surface_points = pn.DataFrame(coords, columns=['X', 'Y', 'Z'])
        surface_points["formation"] = interf_formation
        surface_points["series"] = interf_series

        orientations = pn.DataFrame(grads, columns=['G_x', 'G_y', 'G_z', 'X', 'Y', 'Z'])
        orientations["series"] = orient_series

        dips = []
        azs = []
        pols = []
        for i, row in orientations.iterrows():
            dip, az, pol = gp.data_management.get_orientation((row["G_x"], row["G_y"], row["G_z"]))
            dips.append(dip)
            azs.append(az)
            pols.append(pol)

        orientations["dip"] = dips
        orientations["azimuth"] = azs
        orientations["polarity"] = pols

        return surface_points, orientations


    def get_stratigraphic_column(self):
        """
        Extracts series names from ElementTree root.

        Returns:
            tuple: Series names (str) in stratigraphic order.
        """
        stratigraphic_column = []
        for s in self.get_psc():
            stratigraphic_column.append(s.get("name"))
        return tuple(stratigraphic_column)

    def get_order_formations(self):
        order_formations = []
        for entry in self.series_distribution.values():
            if type(entry) is str:
                order_formations.append(entry)
            elif type(entry) is tuple:
                for e in entry:
                    order_formations.append(e)

        return order_formations

    def get_faults(self):
        """
        Extracts fault names from ElementTree root.

        Returns:
            tuple: Fault names (str) ordered as in the GeoModeller XML.
        """
        faults = []
        for c in self.root[2]:
            faults.append(c.get("Name"))
        return tuple(faults)

    def get_series_distribution(self):
        """
        Combines df and stratigraphic series into an unordered dictionary as keys and maps the correct
        formations to them as a list value. Faults series get a list of their own string assigned as formation.

        Returns:
            (dict): maps Series (str) -> Surfaces (list of str)
        """
        series_distribution = {}
        for key in self.series_info.keys():
            fmts = self.series_info[key]["formations"]
            if len(fmts) == 1:
                series_distribution[key] = fmts[0]
            else:
                series_distribution[key] = tuple(fmts)

        for f in self.stratigraphic_column:
            if "Fault" in f or "fault" in f:
                series_distribution[f] = f

        return series_distribution

    def _get_extent(self):
        """
        Extracts model extent from ElementTree root and returns it as tuple of floats.

        Returns:
            tuple: Model extent as (xmin, xmax, ymin, ymax, zmin, zmax).
        """
        xy = self.root[0][0][0][0].attrib
        z = self.root[0][0][0][1].attrib
        return tuple(np.array([xy["Xmin"], xy["Xmax"],
                               xy["Ymin"], xy["Ymax"],
                                z["Zmin"],  z["Zmax"]]).astype(float))

    def get_surface_points_df(self):
        """
        Extracts the interface data points stored in the GeoModeller xml file and returns it as a GemPy surface_points
        dataframe.

        Returns:
            pandas.DataFrame: InputData.surface_points dataframe
        """
        if self.root.find("{" + self.xmlns + "}Structural3DData") is None:
            print("No 3D data stored in given xml file.")
            return None
        else:
            fmts = [c.attrib["Name"] for c in self.root.find("{" + self.xmlns + "}Structural3DData")[0]]  # use formations
            xyzf = []

            for i, fmt in enumerate(fmts):  # loop over all formations
                for p in self.root[5][0][i]:  # loop over every point
                    entry = p[0].text.split(",")  # split the string by its seperator into coord strings
                    entry.append(fmt)

                    for s in self.series_info.keys():
                        if fmt in self.series_info[s]["formations"]:
                            series = s
                        else:
                            series = fmt

                    entry.append(series)
                    xyzf.append(entry)

            surface_points = pn.DataFrame(np.array(xyzf), columns=['X', 'Y', 'Z', "formation", "series"])
            surface_points[["X", "Y", "Z"]] = surface_points[["X", "Y", "Z"]].astype(float)
            return surface_points

    def get_orientation_df(self):
        """
        Extracts the orientation data points sotred in the GeoModeller xml file and returns it as a GemPy
        orientations dataframe.

        Returns:
            pandas.DataFrame: InputData.orientations dataframe
        """
        if self.root.find("{" + self.xmlns + "}Structural3DData") is None:
            print("No 3D data stored in given xml file.")
            return None
        else:
            fol = []
            for i, s in enumerate(self.root.find("{" + self.xmlns + "}Structural3DData")[1]):
                for c in self.root.find("{" + self.xmlns + "}Structural3DData")[1][i]:
                    entry = c[0][0].text.split(",")
                    entry.append(c.get("Dip"))
                    entry.append(c.get("Azimuth"))
                    # correct polarity from bool str to int
                    pol = c.get("NormalPolarity")
                    if pol == "true":
                        entry.append(1)
                    else:
                        entry.append(-1)
                    entry.append(s.get("Name"))
                    for series in self.series_distribution.keys():
                        if s.get("Name") in self.series_distribution[series]:
                            entry.append(series)

                    fol.append(entry)

            orientations = pn.DataFrame(np.array(fol), columns=['X', 'Y', 'Z', 'dip', 'azimuth', 'polarity', 'formation', 'series'])
            orientations[["X", "Y", "Z", "dip", "azimuth"]] = orientations[["X", "Y", "Z", "dip", "azimuth"]].astype(float)
            orientations["polarity"] = orientations["polarity"].astype(int)

            return orientations

    def _get_series_fmt_dict(self):
        sp = {}
        for i, s in enumerate(self.stratigraphic_column):  # loop over all series
            fmts = []  # init formation storage list
            influenced_by = []  # init influenced by list
            for c in self.root.find("{" + self.xmlns + "}GeologicalModel").find("{"+self.xmlns+"}ProjectStratigraphicColumn")[i]:
                if "Data" in c.tag:
                    fmts.append(c.attrib["Name"])
                elif "InfluencedByFault" in c.tag:
                    influenced_by.append(c.attrib["Name"])
            # print(fmts)
            sp[s] = {}
            sp[s]["formations"] = fmts
            sp[s]["InfluencedByFault"] = influenced_by

        return sp

    def _where_do_faults_stop(self):
        fstop = {}
        for i, f in enumerate(self.root[2]):

            stops_on = []
            for c in self.root[2][i][2:]:
                stops_on.append(c.get("Name"))

            fstop[f.get("Name")] = stops_on

        return fstop

    def get_fault_matrix(self):
        nf = len(self.faults)
        fm = np.zeros((nf, nf))  # zero matrix of n_faults²
        fstop = self._where_do_faults_stop()
        for i, f in enumerate(self.faults):
            for fs in fstop[f]:
                j = np.where(np.array(self.faults) == fs)[0][0]
                fm[i, j] = 1

        return fm


# TODO think where this function should go
def read_vox(geo_data, path):
    """
    read vox from geomodeller and transform it to gempy format
    Returns:
        numpy.array: block model
    """
    import pandas as pn
    geo_res = pn.read_csv(path)

    geo_res = geo_res.iloc[9:]

    # ip_addresses = geo_res['nx 50'].unique()  # geo_model.surface_points["formation"].unique()
    ip_dict = geo_data.get_formation_number()
  #  ip_dict = geo_model.surface_points['formation_number'].unique()

    geo_res_num = geo_res.iloc[:, 0].replace(ip_dict)
    block_geomodeller = np.ravel(geo_res_num.values.reshape(
        geo_data.resolution[0], geo_data.resolution[1], geo_data.resolution[2], order='C').T)
    return block_geomodeller


class GeomodellerClass:
    """Wrapper for GeoModeller XML-datafiles to perform all kinds of data
    manipulation and analysis on low level basis, e.g.:
    - Uncertainty Simulation
    - TWT to depth conversion
    - Data analysis, e.g. Auto-documentation"""

    def __init__(self):
        """Wrapper for GeoModeller XML-datafiles to perform all kinds of data
    manipulation and analysis on low level basis, e.g.:
    - Uncertainty Simulation
    - TWT to depth conversion
    - Data analysis, e.g. Auto-documentation"""

    def load_geomodeller_file(self, xml_file):
        self.xml_file_name = xml_file
        try:
            tree = ET.parse(xml_file)
        except IOError:
            print(("Can not open xml File " + xml_file + ": " + string_err))
            print ("Please check file name and directory and try again")
            raise IOError
        # safe tree on local varibale
        self.tree = tree
        # get rootelement
        self.rootelement = tree.getroot()
        # set other class variables
        self.xmlns = "http://www.geomodeller.com/geo"
        self.gml = "http://www.opengis.net/gml"

    def load_deepcopy_tree(self, deepcopy_tree):
        """load tree information from deepcopied tree into object"""
        self.tree = deepcopy_tree
        self.rootelement = deepcopy_tree.getroot()
        # set other class variables
        self.xmlns = "http://www.geomodeller.com/geo"
        self.gml = "http://www.opengis.net/gml"

    def deepcopy_tree(self):
        """create a deep copy of original tree to restore later, e.g. for uncertainty evaluation"""

        deepcopy_tree = copy.deepcopy(self.tree)
        deepcopy_tree.parent = None
        return deepcopy_tree

    def reload_geomodeller_file(self, deepcopy_tree):
        """restore original tree root from deep copy of orignial tree
        deep copy can be created (not automatically to save memory!) with
        self.deepcopy_tree()
        """
        try:
            self.tree = deepcopy_tree
            self.rootelement = self.tree.getroot()
        except NameError:
            print ("No deep copy of original tree available, please create with self.deepcopy_tree()")

    def get_model_extent(self):
        """get extent of model
        returns (x_min, x_max, y_min, y_max, z_min, z_max)
        and saves extent in self.x_min, self.x_max, etc.
        """
        extent_parent = self.rootelement.find("{"+self.xmlns+"}Extent3DOfProject")
        extentbox3D = extent_parent.find("{"+self.xmlns+"}ExtentBox3D")
        extent3D = extentbox3D.find("{"+self.xmlns+"}Extent3D")
        extent_xy = extent3D.find("{"+self.xmlns+"}ExtentXY")
        extent_z = extent3D.find("{"+self.xmlns+"}ExtentZ")
        self.x_min = float(extent_xy.get("Xmin"))
        self.x_max = float(extent_xy.get("Xmax"))
        self.y_min = float(extent_xy.get("Ymin"))
        self.y_max = float(extent_xy.get("Ymax"))
        self.z_min = float(extent_z.get("Zmin"))
        self.z_max = float(extent_z.get("Zmax"))
        return (self.x_min, self.x_max, self.y_min, self.y_max, self.z_min, self.z_max)

    # def get_model_range(self):
    #     """get model range from model extent, e.g. for automatic mesh generation"""
    #     (x_min, x_max, y_min, y_max, z_min, z_max) = self.get_model_extent()
    #     from numpy import abs
    #     self.range_x = abs(x_max - x_min)
    #     self.range_y = abs(y_max - y_min)
    #     self.range_z = abs(z_max - z_min)
    #     return (self.range_x, self.range_y, self.range_z)

    def get_sections(self):
        """get sections out of rootelement, safe array with section elements
        in local variable"""
        sections_parent = self.rootelement.findall("{"+self.xmlns+"}Sections")[0]
        self.sections = sections_parent.findall("{"+self.xmlns+"}Section")
        return self.sections

    def get_faults(self):
        """get fault elements out of rootelement and safe as local list"""
        try:
            faults_parent = self.rootelement.findall("{"+self.xmlns+"}Faults")[0]
            self.faults = faults_parent.findall("{"+self.xmlns+"}Fault")
        except IndexError:
            print("No df found in model")
        return self.faults

    def get_formations(self):
        """get formation elements out of rootelement and safe as local list"""
        formations_parent = self.rootelement.findall("{"+self.xmlns+"}Surfaces")[0]
        self.formations = formations_parent.findall("{"+self.xmlns+"}Formation")

    def get_stratigraphy_list(self, **kwds):
        """get project stratigraphy and return as list; lowermost formation: 1
        for GeoModeller dll access (this ist the formation_number that is returned with
        the GetComputedLithologyXYZ function in the geomodeller dll
        optional keywords:
        out = string : set 'out' formation to this name (might be necessary for TOUGH2 simulation!)
        """
        series_list = []
        strati_column = self.rootelement.find("{"+self.xmlns+"}GeologicalModel").find("{"+self.xmlns+"}ProjectStratigraphicColumn")#.findall("{"+self.xmlns+"Series")
        series = strati_column.findall("{"+self.xmlns+"}Series")
        for s in series:
            data = s.findall("{"+self.xmlns+"}Data")
            for d in data:
                series_list.append(d.get("Name"))
        # append "out" as uppermost formation for "out values
        if "tough2" in kwds:
            if 'out' in kwds:
                series_list.append(kwds['out'])
            else:
                series_list.append("out")
        self.stratigraphy_list = series_list
        return series_list

    def get_section_names(self):
        """get all section names out of local variable self.sections"""
        # todo: return section names as dictionary with element and name?
        # test if self.sections defined, if not -> create
        try:
            self.sections
        except AttributeError:
            # print "Create sections Data array"
            self.get_sections()
        section_names = []
        for section in self.sections:
            section_names.append(section.get("Name"))
        return section_names

    def get_formation_names(self):
        """get formation names and return as list"""
        forms=[]
        try:
            self.formations
        except AttributeError:
            self.get_formations()
        for formation in self.formations:
            forms.append(formation.get("Name"))
        return forms

    def get_fault_names(self):
        """get fault names and return as list"""
        faults_list=[]
        try:
            self.faults
        except AttributeError:
            self.get_faults()
        for fault in self.faults:
            faults_list.append(fault.get("Name"))
        return faults_list

    def get_points_in_sections(self):
        """Create dictionary of all points (with obs-id) in all sections"""
        self.create_sections_dict()
        for sec in list(self.section_dict.keys()):
            forms = self.get_formation_point_data(self.section_dict[sec])
            if forms == None:
                print ("\t\t\tNo Formation Points in this section")
            else:
                for form in forms:
                    #print form.get("ObservationID")
                #    if form.get("ObservationID") == None: continue
                    data = form.find("{"+self.xmlns+"}Data")
                    print(("\nObsID = %s" % form.get("ObservationID")))
                    print(("\tFormation name\t= %s" % data.get("Name")))
                    element_point = form.find("{"+self.gml+"}LineString")
                    element_coords = element_point.find("{"+self.gml+"}coordinates")
                    tmp = element_coords.text.split(" ")
                    for tmp1 in tmp:
                        if tmp1 == '': continue
                        tmp_cds = tmp1.split(",")
                        print(("\tX = %.1f, Y = %.1f" % (float(tmp_cds[0]), float(tmp_cds[1]))))


                    fol = form.find("{"+self.xmlns+"}FoliationObservation")
                    if fol is not None:
                        print(("\tFoliation defined: azimuth = %.1f, dip = %.1f" % (float(fol.get("Azimuth")), float(fol.get("Dip")))))
                        # get position of foliation (yet another point)
                        pt = fol.find("{"+self.gml+"}Point")
                        c = pt.find("{"+self.gml+"}coordinates")
                        cds = c.text.split(",")
                        print(("\t\tX = %.1f, Y = %.1f" % (float(cds[0]), float(cds[1]))))

            print ("\n")
            print((80*"-"))
            print(("Foliations in section %s:" % sec))
            print((80*"-"))
            foliations = self.get_foliations(self.section_dict[sec])
            if foliations == None:
                print ("\t\t\tNo foliations in this section")
            else:
                for fol1 in foliations:
                    print(("\nObsID = %s" % fol1.get("ObservationID")))
                    data = fol1.find("{"+self.xmlns+"}Data")
                    fol = fol1.find("{"+self.xmlns+"}FoliationObservation")
                    print(( "\tFormation name\t= %s" % data.get("Name")))
                    print(("\tAzimuth = %.1f, dip = %.1f" % (float(fol.get("Azimuth")), float(fol.get("Dip")))))
                    pt = fol.find("{"+self.gml+"}Point")
                    c = pt.find("{"+self.gml+"}coordinates")
                    cds = c.text.split(",")
                    print(("\tX = %.1f, Y = %.1f" % (float(cds[0]), float(cds[1]))))
        return

    def get_formation_parameters(self):
        """read formation parameters; physical
        properties, density, th. cond etc... store in dict"""
        #
        # To do: re-write in a more elegant way and keep original
        # structure and key-words?
        #
        self.formation_params = {}
        try:
            self.formations
        except AttributeError:
            #            print "Create sections Data array"
            self.get_formations()
        for formation in self.formations:
            self.formation_params[formation.get("Name")] = {}
            geophys = formation.find("{"+self.xmlns+"}Geophysics")
            dens = geophys.find("{"+self.xmlns+"}DensityCompoundDistribution")
            dens_simple = dens.find("{"+self.xmlns+"}SimpleDistribution")
            self.formation_params[formation.get("Name")]["dens_mean"] = dens_simple.get("Mean")
            self.formation_params[formation.get("Name")]["dens_law"] = dens_simple.get("LawType")
            self.formation_params[formation.get("Name")]["dens_dev"] = dens_simple.get("Deviation")
            #             print geophys.getchildren()
            mag = geophys.find("{"+self.xmlns+"}RemanantMagnetizationCompoundDistribution")
            mag_simple = mag.find("{"+self.xmlns+"}SimpleDistributionVector")
            self.formation_params[formation.get("Name")]["mag"] = mag_simple.get("Mean")
            velocity = geophys.find("{"+self.xmlns+"}VelocityCompoundDistribution")
            velocity_simple = velocity.find("{"+self.xmlns+"}SimpleDistribution")
            self.formation_params[formation.get("Name")]["vel_mean"] = velocity_simple.get("Mean")
            self.formation_params[formation.get("Name")]["vel_law"] = velocity_simple.get("LawType")
            self.formation_params[formation.get("Name")]["vel_dev"] = velocity_simple.get("Deviation")
            # Thermal properties are only defined in newer versions of GeoModeller! thus check!

            th_cond = geophys.find("{"+self.xmlns+"}ThermalConductivityCompoundDistribution")
            if th_cond == None: continue
            th_cond_simple = th_cond.find("{"+self.xmlns+"}SimpleDistribution")
            self.formation_params[formation.get("Name")]["th_cond_mean"] = th_cond_simple.get("Mean")
            self.formation_params[formation.get("Name")]["th_cond_law"] = th_cond_simple.get("LawType")
            self.formation_params[formation.get("Name")]["th_cond_dev"] = th_cond_simple.get("Deviation")
            heat_prod = geophys.find("{"+self.xmlns+"}HeatProductionRateCompoundDistribution")
            heat_prod_simple = heat_prod.find("{"+self.xmlns+"}SimpleDistribution")
            self.formation_params[formation.get("Name")]["heat_prod_mean"] = heat_prod_simple.get("Mean")
            self.formation_params[formation.get("Name")]["heat_prod_law"] = heat_prod_simple.get("LawType")
            self.formation_params[formation.get("Name")]["heat_prod_dev"] = heat_prod_simple.get("Deviation")

            # same for other properties
        #    print th_cond
            #
            # !!! only simple distributions yet impl.
            #

    def create_fault_dict(self):
        """create dictionary for fault elements with names as keys"""
        # test if self.formations defined, if not -> create
        try:
            self.faults
        except AttributeError:
            print ("Create Surfaces list")
            self.get_faults()
        self.fault_dict = {}
        for fault in self.faults:
            self.fault_dict[fault.get("Name")] = fault
        return self.fault_dict

    def create_formation_dict(self):
        """create dictionary for formation elements with formation names as keys"""
        # test if self.formations defined, if not -> create
        try:
            self.formations
        except AttributeError:
            print ("Create formation dictionary")
            self.get_formations()
        self.formation_dict = {}
        for formation in self.formations:
            self.formation_dict[formation.get("Name")] = formation
        return self.formation_dict

    def create_sections_dict(self):
        """create dictionary for section elements with section names as keys
        (for easier use...)"""
        # test if self.sections defined, if not -> create
        try:
            self.sections
        except AttributeError:
            # print "Create sections dictionary"
            self.get_sections()
        self.section_dict = {}
        for section in self.sections:
            self.section_dict[section.get("Name")] = section
        return self.section_dict

    def get_foliations(self, section_element):
        """get all foliation data elements from a for section"""
        tmp_element = section_element.find("{"+self.xmlns+"}Structural2DData")
        # check in case there is no foliation defined in this section
        # tmp_element2 = tmp_element.find("{"+self.xmlns+"}Foliations")
        try:
            tmp_element2 = tmp_element.find("{"+self.xmlns+"}Foliations")
        except AttributeError:
            return None
        try:
            foliations = tmp_element2.findall("{"+self.xmlns+"}Foliation")
        except AttributeError:
            return None
        return foliations

    def get_foliation_dip(self, foliation_element):
        """get dip of foliation element"""
        return float(foliation_element.find("{"+self.xmlns+"}FoliationObservation").get("Dip"))

    def get_foliation_azimuth(self, foliation_element):
        """get dip of foliation element"""
        return float(foliation_element.find("{"+self.xmlns+"}FoliationObservation").get("Azimuth"))

    def get_folation_polarity(self, foliation_element):
        """get polarity of foliation element; return true if Normal Polarity"""
        return foliation_element.find("{"+self.xmlns+"}FoliationObservation").get("NormalPolarity")

    def get_foliation_coordinates(self, foliation_element):
        """get coordinates of foliation element"""
        element_fol = foliation_element.find("{"+self.xmlns+"}FoliationObservation")
        element_point = element_fol.find("{"+self.gml+"}Point")
        element_coords = element_point.find("{"+self.gml+"}coordinates")
        return str(element_coords.text)

    def get_formation_data(self, section_element):
        """not used any more! use get_formation_point_data(section_element) instead"""
        print ("not used any more! use get_formation_point_data(section_element) instead")
        return None

    def get_formation_point_data(self, section_element):
        """get all formation point data elements from a for section"""
        tmp_element = section_element.find("{"+self.xmlns+"}Structural2DData")
        # check in case there is no formation points defined in this section
        try:
            tmp_element2 = tmp_element.find("{"+self.xmlns+"}SurfacePoints")
        except AttributeError:
            return None
        return tmp_element2.findall("{"+self.xmlns+"}Interface")

    def get_name(self, section_element):
        """get the name of any section element (if defined)"""
        return section_element.find("{"+self.xmlns+"}Name")

    def get_interface_name(self, interface_element):
        """get name of interface, i.e. the formation"""
        return interface_element.find("{"+self.xmlns+"}Data").get("Name")

    def get_point_coordinates(self, point_elements, **args):
        """get the coordinates of a specific point memory locations"""
        point_list = list()

        for element in point_elements:
          name = element.find("{"+self.xmlns+"}Data").get("Name")
          #if args.has_key("if_name"):
          if "if_name" in args:
            if args["if_name"] != name: continue
          element_point = element.find("{"+self.gml+"}LineString")
          element_coords = element_point.find("{"+self.gml+"}coordinates")
          point_list.append((name+ " " + str(element_coords.text)))
        return point_list

    def change_formation_values_PyMC(self, **args):
        """ -So far is ready only to changes points in coordinates y. It is not difficult to add a new
        dimension

            - The dips and azimuth ObservationID must contain _d or _a respectively"""

        if "info" in args:
            section_dict = self.create_sections_dict()
            contact_points_dict = {}
            foliation_dict = {}
            for i in range(len(section_dict)):
                print(("\n\n\n", list(section_dict.keys())[i], "\n"))
                print ("Elements and their ID \n")
                contact_points = self.get_formation_point_data(list(section_dict.values())[i])

                try:
                    for contact_point in contact_points:
                        contact_points_dict[contact_point.get("ObservationID")] = contact_point
                        print((contact_point, contact_point.get("ObservationID")))
                except TypeError:
                    print ("No contact points in the section")
                #ObsID = contact_points.get("ObservationID")
                foliations = self.get_foliations(list(section_dict.values())[i])
                try:
                    for foliation in foliations:
                        # dictionary to access with azimth name
                        foliation_dict[foliation.get("ObservationID")+"_a"] = foliation
                        # dictionary to access with dip name
                        foliation_dict[foliation.get("ObservationID")+"_d"] = foliation
                        print((foliation, foliation.get("ObservationID")))

                except TypeError:
                    print ("No foliation in the section")
                try:
                    coord_interface = self.get_point_coordinates(contact_points)
                except TypeError:
                    print ("Element does not have iterable objects")

                print(("\nDictionaries:\n ", contact_points_dict, "\n", foliation_dict))

                print(("\n Contact points", contact_points, "\n", coord_interface, "\n"))

                print(("foliations" , foliations,  "\n"))
                try:
                    for i in range(len(foliations)):
                        print(("azimut:",self.get_foliation_azimuth(foliations[i])))
                        print(("dip",self.get_foliation_dip(foliations[i])))
                        print(("coordinates", self.get_foliation_coordinates(foliations[i])))
                except TypeError:
                    print ("No foliation in the section")
            return None
        #========================
        # change the stuff
        #=======================
        section_dict = self.create_sections_dict()
        contact_points_dict = {}
        foliation_dict = {}

        #Creation of dictionaries according to the ObservationID
        for i in range(len(section_dict)):
            # Contact points:
            try:
                contact_points = self.get_formation_point_data(list(section_dict.values())[i])
                for contact_point in contact_points:
                    contact_points_dict[contact_point.get("ObservationID")] = contact_point
            except TypeError:
                continue
            # Foliation Points
            try:
                foliations = self.get_foliations(list(section_dict.values())[i])
                for foliation in foliations:
                    # dictionary to access with azimth name
                    foliation_dict[foliation.get("ObservationID")+"_a"] = foliation
                    # dictionary to access with dip name
                    foliation_dict[foliation.get("ObservationID")+"_d"] = foliation
            except TypeError:
                continue

        # Passing our chain values:
            # Contact_points
        if "contact_points_mc" in args:
            for contac_point_mc in args["contact_points_mc"]:
                try:

                    element = contact_points_dict[str(contac_point_mc)]
                    element_point = element.find("{"+self.gml+"}LineString")
                    element_coords = element_point.find("{"+self.gml+"}coordinates")
                    point_list = element_coords.text.split(" ")
                    if point_list[-1] == '':
                        point_list = point_list[0:-1]

                    if len(point_list) == 1:
                        self.change_formation_point_pos(element, y_coord = contac_point_mc.value)
                    #Specific case of the Graben:
                    elif len(point_list) == 2:
                        self.change_formation_point_pos(element, y_coord = [contac_point_mc.value, contac_point_mc.value])
                    else:
                        print ("The lenght of the points to change does not fit with the number of changes in the input (>2)")
                except KeyError:
                    print(("The name of your PyMC variables (%s) does not agree with the ObservationID in the xml. Check misspellings." % str(contac_point_mc)))
                    continue
            # Azimuths
        if "azimuths_mc" in args:
            for azimuth_mc in args["azimuths_mc"]:
                #print azimuth_mc, type(azimuth_mc)
                try:
                    self.change_foliation(foliation_dict[str(azimuth_mc)], azimuth = str(azimuth_mc.value))
                except KeyError:
                    print(("The name of your PyMC variables (%s) does not agree with the ObservationID in the xml. Check misspellings." % str(azimuth_mc)))
                    continue
            # Dips
        if "dips_mc" in args:
            for dip_mc in args["dips_mc"]:
                try:
                    self.change_foliation(foliation_dict[str(dip_mc)], dip = str(dip_mc.value))
                except KeyError:
                    print(("The name of your PyMC variables (%s) does not agree with the ObservationID in the xml. Check misspellings." % str(dip_mc)))
                    continue


    # To do: vectorize this
    def change_formation_point_pos(self, element, **args):
        """change position of formation point in section element
        arguments:
        x_coord, y_coord : set to this coordinates
        add_x_coord, add_y_coord : add values to existing coordinates
        use if_name = and if_provenance = to add conditions!
        print_points = bool: print the list of points that will be modified (default: False)"""
        #    print "I am here"
        #print_points = kwds.get('print_points', False)
        prov = element.get("Provenance")

        name = element.find("{"+self.xmlns+"}Data").get("Name")

        #if args.has_key("if_name"):
        if "if_name" in args:
            if args["if_name"] != name: return
        # if args.has_key("if_provenance"):
        if "if_provenance" in args:
            if args["if_provenance"] != prov: return
        # element_fol = element.find("{"+self.xmlns+"}")
        element_point = element.find("{"+self.gml+"}LineString")
        element_coords = element_point.find("{"+self.gml+"}coordinates")
        point_list = element_coords.text.split(" ")
        #    print "poitn lits", point_list
        if point_list[-1] == '':
            point_list = point_list[0:-1]
        if len(point_list) > 1:
            x_coords = []
            y_coords = []
            if "print_points" in args:
                print (point_list)
            for point in point_list:
                # if point == '': continue
                a = point.split(',')
                #print a
                [x_coord, y_coord] = [float(a[0]), float(a[1])]
                x_coords.append(x_coord)
                y_coords.append(y_coord)
            # convert to arrays for calculation
            x_coords = array(x_coords)
            y_coords = array(y_coords)
            # Here  y_coord, and x_coord
            if "x_coord" in args:
                if shape(point_list) == shape(args["x_coord"]):
                #except TypeError:
                    x_coords = array(args["x_coord"])
                else:
                    print ("length of the points you want to change do not match with input dimensions")
            if "y_coord" in args:
                #print (args["y_coord"])
                #print array(args["y_coord"])
                if shape(point_list) == shape(args["y_coord"]):
                    y_coords = array(args["y_coord"])
                    #            print "ycoords", y_coords
                else:
                    print ("length of the points you want to change do not match with input dimensions")
            #print "Coordenates", x_coords, y_coords
            # Here add coords
            if "add_x_coord" in args:
                x_coords = x_coords + float(args["add_x_coord"])
            if "add_y_coord" in args:
                y_coords = y_coords + float(args["add_y_coord"])
            #    print y_coords
            # now, reconstruct output format strings
            out_text = ''
            for (i, x_coord) in enumerate(x_coords):
                out_text += "%f,%f " % (x_coords[i],y_coords[i])
            element_coords.text = out_text

        else:
            [x_coord, y_coord] = point_list[0].split(",")
            [x_coord, y_coord] = [float(x_coord), float(y_coord)]
            if "x_coord" in args:
                x_coord = float(args["x_coord"])
            if "y_coord" in args:
                y_coord = float(args["y_coord"])
            if "add_x_coord" in args:
                x_coord = x_coord + float(args["add_x_coord"])
            if "add_y_coord" in args:
                y_coord = y_coord + float(args["add_y_coord"])
            element_coords.text = "%f,%f" % (x_coord, y_coord)
        return None

    def change_foliation_polarity(self, element):
        """change polarity of foliation element"""
        if element.get("NormalPolarity") == "true":
            element.set("NormalPolarity", "false")
        else:
            element.set("NormalPolarity", "true")

    def change_foliation(self, element, **args):
        """change foliation data, argument one or more of: azimuth, dip,
        normalpolarity = true/false, x_coord, y_coord" or: add_dip, add_azimuth,
        add_x_coord, add_y_coord to add values to existing values!
        use if_name = and if_provenance = to add conditions!"""
        prov = element.get("Provenance")
        name = element.find("{"+self.xmlns+"}Data").get("Name")
        if "if_name" in args:
            if args["if_name"] != name: return
        if "if_provenance" in args:
            if args["if_provenance"] != prov: return
        element_fol = element.find("{"+self.xmlns+"}FoliationObservation")
        if "dip" in args:
            element_fol.set("Dip", args["dip"])
        if "azimuth" in args:
            element_fol.set("Azimuth", args["azimuth"])
        if "nomalpolarity" in args:
            element_fol.set("NormalPolarity", args["normalpolarity"])
        #
        # To Do: logfile, if dip >90, azi > 360, ...
        #
        if "add_dip" in args:
            dip_org = float(element_fol.get("Dip"))
            dip_new = dip_org + float(args["add_dip"])
            if dip_new > 90:
                dip_new = 180 - dip_new
                self.change_foliation_polarity(element_fol)
                azi_org = float(element_fol.get("Azimuth"))
                if azi_org < 180:
                    element_fol.set("Azimuth", str(azi_org+180))
                else:
                    element_fol.set("Azimuth", str(azi_org-180))
            element_fol.set("Dip", str(dip_new))
        if "add_azimuth" in args:
            azi_org = float(element_fol.get("Azimuth"))
            azi_new = azi_org + float(args["add_azimuth"])
            if azi_new > 360.0: azi_new -= 360
            element_fol.set("Azimuth", str(azi_new))
        element_point = element_fol.find("{"+self.gml+"}Point")
        element_coords = element_point.find("{"+self.gml+"}coordinates")
        [x_coord, y_coord] = element_coords.text.split(",")
        [x_coord, y_coord] = [float(x_coord), float(y_coord)]
        if "x_coord" in args:
            x_coord = float(args["x_coord"])
        if "y_coord" in args:
            y_coord = float(args["y_coord"])
        if "add_x_coord" in args:
            x_coord = x_coord + float(args["add_x_coord"])
        if "add_y_coord" in args:
            y_coord = y_coord + float(args["add_y_coord"])
        element_coords.text = "%f,%f" % (x_coord, y_coord)
        return None

    def twt_to_depth(self, sec_element, formula, **args):
        """Convert all data within a section from twt to depth (including
        orientation data!!
        Input: section element with data points, conversion function as
        string with 't' as placeholder for twt-time, e.g. '2 * (-t) ** 2 + 18 * (-t)'
        ATTENTION: check if t negative
        optional arguments:
        change_dip (boolean) : change dip angle in foliation data
        according to first derivative of twt-to-depth formula
        create_plot (boolean) : create summary plot with twt and converted depth
        for conversion formula control
        """
        # Idea: stoachstic apporach to twt -> depth conversion: apply several
        # possible formulae for quality estimation of resulting model?
        struct_data = sec_element.find("{"+self.xmlns+"}Structural2DData")
        surface_points = struct_data.find("{"+self.xmlns+"}SurfacePoints").findall("{"+self.xmlns+"}Interface")
        # save data in list to create a plot to check validity of conversion
        t_list = []
        v_list = []
        for interface in surface_points:
            gml_coords_element = interface.find("{"+self.gml+"}LineString").find("{"+self.gml+"}coordinates")
            # check for correct decimal, column (cs) and text separator (ts)
            ts = gml_coords_element.get("ts")
            cs = gml_coords_element.get("cs")
            data_array = gml_coords_element.text.split(ts)
            # check if last entry is empty (seems to happen often), if so: delete!
            # print gml_coords_element.text
            if data_array[-1] == "": del data_array[-1]
            text_new = ""
            # apply conversion formula for z-direction, i.e. dv;
            # no change in x-direction -> du = 0 (but: maybe neccessary for specific situations?)
            for entry in data_array:
                parts = entry.split(cs)
                # get original values
                # t as varibale, as defined in formula (input)
                t = float(parts[1])
                v_new = eval(formula)
                du = 0
                text_new += "%f%s%f%s" % (float(parts[0])+du, cs, v_new, ts)
                # append to list for check-plot
                t_list.append(-t)
                v_list.append(v_new)
            # print text_new
            gml_coords_element.text = text_new
            # print gml_coords_element.text
        # now change foliation position and dip angle! (check: given as argument?)
        # for dip angle: numerical determination of first derivative for
        # twt to depth conversion formula?
        if "change_dip" in args and args["change_dip"]:
            print ("change dip in seismic profile")

        # create check-plot
        if "create_plot" in args and args["create_plot"]:
            print ("Create plot with twt, converted depth pairs")
            plot(t_list,v_list,'.', label = formula)
            title("TWT to depth: Converted data points\nSection: " + sec_element.get("Name"))
            xlabel("TWT [ms]")
            ylabel("Converted depth [m]")
            legend()
            grid(True)
            show()

    def get_pole_points(self, element):
        # function to plot data points in geomodeller element
        u = []
        v = []
        poles = element.getiterator("{"+self.xmlns+"}Pole-Weight")
        for pole in poles:
            u.append(pole.get("U"))
            v.append(pole.get("V"))

        return (u,v)

    def plot_points(self, element):
        # plot u,v points in simple 2D plot
        (u,v) = self.get_pole_points(element)
        plot(u,v,'o')
        name = element.get("Name")
        title(name)
        savefig(name+"_points.png")

    def write_xml(self, save_dir, **args):
        """Write elementtree to xml-file
        arguments:
        print_path: Print the path where the xml is created"""
        # to do: filename (and directory?) as optional argument"""
        # flo, 10/2008
        #file = "changed_model.xml"
        tree_new = ET.ElementTree(self.rootelement)
        if "print_path" in args:
            print(("Write tree to file " + save_dir))
        tree_new.write(save_dir)
        #self.tree.write("changed_model.xml")
        self.tree.write(save_dir)

    def add_to_project_name(self, s):
        """add string s to project name, e.g. Number of uncertainty run"""
        name = self.rootelement.get("projectName")
        name_new = name + " " + s
        self.rootelement.set("projectName",name_new)

    def create_TOUGH_formation_names(self, **kwds):
        """create formation names that are compatible with format required by TOUGH,
        i.e. String of length 5
        returns and stores as dictionary with Geomodeller Names as key and TOUGH names as entry
        (self.tough_formation_names)
        simply cuts original formation name to a name length of 5;
        if cut name already exists: create new name, three first string values followed by two integers
        that are subsequently increased
        optional keywods:
        out = string : set out formation to this name (for TOUGH2: no leading spaces allowed! set to 5 chars!)
        """
        # import str
        self.tough_formation_names = {}
        # check if self.formation_names list already exists, if not: create
        try: self.formation_names
        except AttributeError:
            self.formation_names = self.get_stratigraphy_list(**kwds)
        # create list with tough names to check if name already exists
        tough_name_list = []
        for i,name in enumerate(self.formation_names):
            #    if self.formation_names[i] == '  out' or self.formation_names[i] == '  OUT':
            #   tough_name_list.append("OUT  ")
            #   continue
            cut_name = self.formation_names[i][0:5]
            if cut_name in tough_name_list:
                for j in range(100):
                    if "%3s%02d" % (cut_name[0:3],j) in tough_name_list:
                        continue
                    else:
                        cut_name = "%3s%02d" % (cut_name[0:3],j)
                        tough_name_list.append(cut_name)
                        break
            else:
                tough_name_list.append(cut_name)
            self.tough_formation_names[name] = "%5s" % str.upper(cut_name)
        return self.tough_formation_names


if __name__ == '__main__':
    print ("main")



"""
Model 5 - Fault
===============

"""

# %%
# A simple fault model with constant offset. We start by importing the
# necessary dependencies:
# 

# %%
# Importing GemPy
import gempy_legacy as gp

import pandas as pd
pd.set_option('display.precision', 2)

# %%
# Creating the model by importing the input data and displaying it:
# 

# %%
data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'
path_to_data = data_path + "/data/input_data/jan_models/"

geo_data = gp.create_data('fault', extent=[0, 1000, 0, 1000, 0, 1000], resolution=[50, 50, 50],
                          path_o=path_to_data + "model5_orientations.csv",
                          path_i=path_to_data + "model5_surface_points.csv")

# %% 
geo_data.get_data()

# %%
# Setting and ordering the units and series:
# 

# %% 
gp.map_stack_to_surfaces(geo_data, {"Fault_Series": 'fault',
                                    "Strat_Series": ('rock2', 'rock1')})
geo_data.set_is_fault(['Fault_Series'])

# %%
gp.plot_2d(geo_data, direction='y')

# %%
# Calculating the model:
# 

# %% 
interp_data = gp.set_interpolator(geo_data, aesara_optimizer='fast_compile')

# %% 
sol = gp.compute_model(geo_data)

# %%
# Displaying the result in x and y direction:
# 

# %%
# sphinx_gallery_thumbnail_number = 2
gp.plot_2d(geo_data, cell_number=25,
           direction='y', show_data=False, show_all_data=True)

# %%
gp.plot_2d(geo_data, cell_number=25,
           direction='x', show_data=True)

# %% 

gp.plot_2d(geo_data, cell_number=25,
           direction='y', show_data=True, show_scalar=True, series_n=1)


gp.save_model(geo_data)